import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3 import (
    Angle,
    BesselOption,
    DcSysTime,
    Drive,
    EmitIntensity,
    FixedCompletionSteps,
    FixedUpdateRate,
    FocusOption,
    GainSTMOption,
    GPIOIn,
    Phase,
    PlaneOption,
    Segment,
    SilencerTarget,
)
from pyautd3.native_methods.autd3capi_driver import (
    ControllerPtr,
    DatagramPtr,
    DebugTypeWrap,
    DevicePtr,
    Duration,
    FociSTMPtr,
    GainPtr,
    GainSTMPtr,
    GeometryPtr,
    LinkPtr,
    LoopBehavior,
    ModulationPtr,
    OptionDuration,
    ResultDuration,
    ResultF32,
    ResultSamplingConfig,
    ResultStatus,
    ResultU16,
    SamplingConfigWrap,
    SenderPtr,
    SleeperWrap,
    TransducerPtr,
    TransitionModeWrap,
)
from pyautd3.native_methods.structs import Point3, Quaternion, Vector3


class FPGAStateListPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FPGAStateListPtr) and self._fields_ == other._fields_  # pragma: no cover


class FixedCompletionTime(ctypes.Structure):
    _fields_ = [("intensity", Duration), ("phase", Duration), ("strict_mode", ctypes.c_bool)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FixedCompletionTime) and self._fields_ == other._fields_  # pragma: no cover


class GroupGainMapPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GroupGainMapPtr) and self._fields_ == other._fields_  # pragma: no cover


class ModulationCachePtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ModulationCachePtr) and self._fields_ == other._fields_  # pragma: no cover


class FourierOption(ctypes.Structure):
    _fields_ = [("has_scale_factor", ctypes.c_bool), ("scale_factor", ctypes.c_float), ("clamp", ctypes.c_bool), ("offset", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FourierOption) and self._fields_ == other._fields_  # pragma: no cover


class SquareOption(ctypes.Structure):
    _fields_ = [("low", ctypes.c_uint8), ("high", ctypes.c_uint8), ("duty", ctypes.c_float), ("sampling_config_div", ctypes.c_uint16)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SquareOption) and self._fields_ == other._fields_  # pragma: no cover


class SineOption(ctypes.Structure):
    _fields_ = [
        ("intensity", ctypes.c_uint8),
        ("offset", ctypes.c_uint8),
        ("phase", Angle),
        ("clamp", ctypes.c_bool),
        ("sampling_config_div", ctypes.c_uint16),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SineOption) and self._fields_ == other._fields_  # pragma: no cover


class ResultFPGAStateList(ctypes.Structure):
    _fields_ = [("result", FPGAStateListPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultFPGAStateList) and self._fields_ == other._fields_  # pragma: no cover


class ResultController(ctypes.Structure):
    _fields_ = [("result", ControllerPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultController) and self._fields_ == other._fields_  # pragma: no cover


class FirmwareVersionListPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, FirmwareVersionListPtr) and self._fields_ == other._fields_  # pragma: no cover


class SenderOption(ctypes.Structure):
    _fields_ = [
        ("send_interval", Duration),
        ("receive_interval", Duration),
        ("timeout", OptionDuration),
        ("parallel", ctypes.c_uint8),
        ("sleeper", SleeperWrap),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SenderOption) and self._fields_ == other._fields_  # pragma: no cover


class ResultFirmwareVersionList(ctypes.Structure):
    _fields_ = [("result", FirmwareVersionListPtr), ("err_len", ctypes.c_uint32), ("err", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ResultFirmwareVersionList) and self._fields_ == other._fields_  # pragma: no cover


class GainCachePtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GainCachePtr) and self._fields_ == other._fields_  # pragma: no cover


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi{bin_ext}"))

        self.dll.AUTDControllerGroup.argtypes = [
            SenderPtr,
            ctypes.c_void_p,
            ctypes.c_void_p,
            GeometryPtr,
            ctypes.POINTER(ctypes.c_int32),
            ctypes.POINTER(DatagramPtr),
            ctypes.c_uint16,
        ]
        self.dll.AUTDControllerGroup.restype = ResultStatus

        self.dll.AUTDControllerOpen.argtypes = [ctypes.POINTER(Point3), ctypes.POINTER(Quaternion), ctypes.c_uint16, LinkPtr, SenderOption]
        self.dll.AUTDControllerOpen.restype = ResultController

        self.dll.AUTDControllerClose.argtypes = [ControllerPtr]
        self.dll.AUTDControllerClose.restype = ResultStatus

        self.dll.AUTDControllerFPGAState.argtypes = [ControllerPtr]
        self.dll.AUTDControllerFPGAState.restype = ResultFPGAStateList

        self.dll.AUTDControllerFPGAStateGet.argtypes = [FPGAStateListPtr, ctypes.c_uint32]
        self.dll.AUTDControllerFPGAStateGet.restype = ctypes.c_int16

        self.dll.AUTDControllerFPGAStateDelete.argtypes = [FPGAStateListPtr]
        self.dll.AUTDControllerFPGAStateDelete.restype = None

        self.dll.AUTDControllerFirmwareVersionListPointer.argtypes = [ControllerPtr]
        self.dll.AUTDControllerFirmwareVersionListPointer.restype = ResultFirmwareVersionList

        self.dll.AUTDControllerFirmwareVersionGet.argtypes = [FirmwareVersionListPtr, ctypes.c_uint32, ctypes.c_char_p]
        self.dll.AUTDControllerFirmwareVersionGet.restype = None

        self.dll.AUTDControllerFirmwareVersionListPointerDelete.argtypes = [FirmwareVersionListPtr]
        self.dll.AUTDControllerFirmwareVersionListPointerDelete.restype = None

        self.dll.AUTDFirmwareLatest.argtypes = [ctypes.c_char_p]
        self.dll.AUTDFirmwareLatest.restype = None

        self.dll.AUTDSender.argtypes = [ControllerPtr, SenderOption]
        self.dll.AUTDSender.restype = SenderPtr

        self.dll.AUTDSenderSend.argtypes = [SenderPtr, DatagramPtr]
        self.dll.AUTDSenderSend.restype = ResultStatus

        self.dll.AUTDSpinSleepDefaultAccuracy.argtypes = []
        self.dll.AUTDSpinSleepDefaultAccuracy.restype = ctypes.c_uint32

        self.dll.AUTDSenderOptionIsDefault.argtypes = [SenderOption]
        self.dll.AUTDSenderOptionIsDefault.restype = ctypes.c_bool

        self.dll.AUTDDatagramClear.argtypes = []
        self.dll.AUTDDatagramClear.restype = DatagramPtr

        self.dll.AUTDDatagramDebugSettings.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDDatagramDebugSettings.restype = DatagramPtr

        self.dll.AUTDDatagramForceFan.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDDatagramForceFan.restype = DatagramPtr

        self.dll.AUTDDatagramTuple.argtypes = [DatagramPtr, DatagramPtr]
        self.dll.AUTDDatagramTuple.restype = DatagramPtr

        self.dll.AUTDDatagramPhaseCorr.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDDatagramPhaseCorr.restype = DatagramPtr

        self.dll.AUTDDatagramPulseWidthEncoder.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDDatagramPulseWidthEncoder.restype = DatagramPtr

        self.dll.AUTDDatagramPulseWidthEncoderDefault.argtypes = []
        self.dll.AUTDDatagramPulseWidthEncoderDefault.restype = DatagramPtr

        self.dll.AUTDDatagramReadsFPGAState.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDDatagramReadsFPGAState.restype = DatagramPtr

        self.dll.AUTDDatagramSwapSegmentModulation.argtypes = [ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDDatagramSwapSegmentModulation.restype = DatagramPtr

        self.dll.AUTDDatagramSwapSegmentFociSTM.argtypes = [ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDDatagramSwapSegmentFociSTM.restype = DatagramPtr

        self.dll.AUTDDatagramSwapSegmentGainSTM.argtypes = [ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDDatagramSwapSegmentGainSTM.restype = DatagramPtr

        self.dll.AUTDDatagramSwapSegmentGain.argtypes = [ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDDatagramSwapSegmentGain.restype = DatagramPtr

        self.dll.AUTDDatagramSilencerFromUpdateRate.argtypes = [FixedUpdateRate, ctypes.c_uint8]
        self.dll.AUTDDatagramSilencerFromUpdateRate.restype = DatagramPtr

        self.dll.AUTDDatagramSilencerFromCompletionSteps.argtypes = [FixedCompletionSteps, ctypes.c_uint8]
        self.dll.AUTDDatagramSilencerFromCompletionSteps.restype = DatagramPtr

        self.dll.AUTDDatagramSilencerFromCompletionTime.argtypes = [FixedCompletionTime, ctypes.c_uint8]
        self.dll.AUTDDatagramSilencerFromCompletionTime.restype = DatagramPtr

        self.dll.AUTDDatagramSilencerFixedCompletionStepsIsDefault.argtypes = [FixedCompletionSteps, ctypes.c_uint8]
        self.dll.AUTDDatagramSilencerFixedCompletionStepsIsDefault.restype = ctypes.c_bool

        self.dll.AUTDSTMFoci.argtypes = [SamplingConfigWrap, ctypes.c_void_p, ctypes.c_uint16, ctypes.c_uint8]
        self.dll.AUTDSTMFoci.restype = FociSTMPtr

        self.dll.AUTDSTMFociIntoDatagramWithSegment.argtypes = [FociSTMPtr, ctypes.c_uint8, ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDSTMFociIntoDatagramWithSegment.restype = DatagramPtr

        self.dll.AUTDSTMFociIntoDatagramWithLoopBehavior.argtypes = [FociSTMPtr, ctypes.c_uint8, ctypes.c_uint8, TransitionModeWrap, LoopBehavior]
        self.dll.AUTDSTMFociIntoDatagramWithLoopBehavior.restype = DatagramPtr

        self.dll.AUTDSTMFociIntoDatagram.argtypes = [FociSTMPtr, ctypes.c_uint8]
        self.dll.AUTDSTMFociIntoDatagram.restype = DatagramPtr

        self.dll.AUTDSTMGain.argtypes = [SamplingConfigWrap, ctypes.POINTER(GainPtr), ctypes.c_uint16, GainSTMOption]
        self.dll.AUTDSTMGain.restype = GainSTMPtr

        self.dll.AUTDSTMGainIntoDatagramWithSegment.argtypes = [GainSTMPtr, ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDSTMGainIntoDatagramWithSegment.restype = DatagramPtr

        self.dll.AUTDSTMGainIntoDatagramWithLoopBehavior.argtypes = [GainSTMPtr, ctypes.c_uint8, TransitionModeWrap, LoopBehavior]
        self.dll.AUTDSTMGainIntoDatagramWithLoopBehavior.restype = DatagramPtr

        self.dll.AUTDSTMGainIntoDatagram.argtypes = [GainSTMPtr]
        self.dll.AUTDSTMGainIntoDatagram.restype = DatagramPtr

        self.dll.AUTDSTMConfigFromFreq.argtypes = [ctypes.c_float, ctypes.c_uint16]
        self.dll.AUTDSTMConfigFromFreq.restype = ResultSamplingConfig

        self.dll.AUTDSTMConfigFromPeriod.argtypes = [Duration, ctypes.c_uint16]
        self.dll.AUTDSTMConfigFromPeriod.restype = ResultSamplingConfig

        self.dll.AUTDSTMConfigFromFreqNearest.argtypes = [ctypes.c_float, ctypes.c_uint16]
        self.dll.AUTDSTMConfigFromFreqNearest.restype = SamplingConfigWrap

        self.dll.AUTDSTMConfigFromPeriodNearest.argtypes = [Duration, ctypes.c_uint16]
        self.dll.AUTDSTMConfigFromPeriodNearest.restype = SamplingConfigWrap

        self.dll.AUTDDatagramSynchronize.argtypes = []
        self.dll.AUTDDatagramSynchronize.restype = DatagramPtr

        self.dll.AUTDDcSysTimeNow.argtypes = []
        self.dll.AUTDDcSysTimeNow.restype = DcSysTime

        self.dll.AUTDDebugTypeNone.argtypes = []
        self.dll.AUTDDebugTypeNone.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeBaseSignal.argtypes = []
        self.dll.AUTDDebugTypeBaseSignal.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeThermo.argtypes = []
        self.dll.AUTDDebugTypeThermo.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeForceFan.argtypes = []
        self.dll.AUTDDebugTypeForceFan.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeSync.argtypes = []
        self.dll.AUTDDebugTypeSync.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeModSegment.argtypes = []
        self.dll.AUTDDebugTypeModSegment.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeModIdx.argtypes = [ctypes.c_uint16]
        self.dll.AUTDDebugTypeModIdx.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeStmSegment.argtypes = []
        self.dll.AUTDDebugTypeStmSegment.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeStmIdx.argtypes = [ctypes.c_uint16]
        self.dll.AUTDDebugTypeStmIdx.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeIsStmMode.argtypes = []
        self.dll.AUTDDebugTypeIsStmMode.restype = DebugTypeWrap

        self.dll.AUTDDebugTypePwmOut.argtypes = [TransducerPtr]
        self.dll.AUTDDebugTypePwmOut.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeDirect.argtypes = [ctypes.c_bool]
        self.dll.AUTDDebugTypeDirect.restype = DebugTypeWrap

        self.dll.AUTDDebugTypeSysTimeEq.argtypes = [DcSysTime]
        self.dll.AUTDDebugTypeSysTimeEq.restype = DebugTypeWrap

        self.dll.AUTDLoopBehaviorInfinite.argtypes = []
        self.dll.AUTDLoopBehaviorInfinite.restype = LoopBehavior

        self.dll.AUTDLoopBehaviorFinite.argtypes = [ctypes.c_uint16]
        self.dll.AUTDLoopBehaviorFinite.restype = LoopBehavior

        self.dll.AUTDPhaseFromRad.argtypes = [ctypes.c_float]
        self.dll.AUTDPhaseFromRad.restype = ctypes.c_uint8

        self.dll.AUTDPhaseToRad.argtypes = [Phase]
        self.dll.AUTDPhaseToRad.restype = ctypes.c_float

        self.dll.AUTDSamplingConfigFromDivision.argtypes = [ctypes.c_uint16]
        self.dll.AUTDSamplingConfigFromDivision.restype = ResultSamplingConfig

        self.dll.AUTDSamplingConfigFromFreq.argtypes = [ctypes.c_float]
        self.dll.AUTDSamplingConfigFromFreq.restype = SamplingConfigWrap

        self.dll.AUTDSamplingConfigFromPeriod.argtypes = [Duration]
        self.dll.AUTDSamplingConfigFromPeriod.restype = SamplingConfigWrap

        self.dll.AUTDSamplingConfigIntoNearest.argtypes = [SamplingConfigWrap]
        self.dll.AUTDSamplingConfigIntoNearest.restype = SamplingConfigWrap

        self.dll.AUTDSamplingConfigDivision.argtypes = [SamplingConfigWrap]
        self.dll.AUTDSamplingConfigDivision.restype = ResultU16

        self.dll.AUTDSamplingConfigFreq.argtypes = [SamplingConfigWrap]
        self.dll.AUTDSamplingConfigFreq.restype = ResultF32

        self.dll.AUTDSamplingConfigPeriod.argtypes = [SamplingConfigWrap]
        self.dll.AUTDSamplingConfigPeriod.restype = ResultDuration

        self.dll.AUTDSamplingConfigEq.argtypes = [SamplingConfigWrap, SamplingConfigWrap]
        self.dll.AUTDSamplingConfigEq.restype = ctypes.c_bool

        self.dll.AUTDTransitionModeSyncIdx.argtypes = []
        self.dll.AUTDTransitionModeSyncIdx.restype = TransitionModeWrap

        self.dll.AUTDTransitionModeSysTime.argtypes = [DcSysTime]
        self.dll.AUTDTransitionModeSysTime.restype = TransitionModeWrap

        self.dll.AUTDTransitionModeGPIO.argtypes = [ctypes.c_uint8]
        self.dll.AUTDTransitionModeGPIO.restype = TransitionModeWrap

        self.dll.AUTDTransitionModeExt.argtypes = []
        self.dll.AUTDTransitionModeExt.restype = TransitionModeWrap

        self.dll.AUTDTransitionModeImmediate.argtypes = []
        self.dll.AUTDTransitionModeImmediate.restype = TransitionModeWrap

        self.dll.AUTDTransitionModeNone.argtypes = []
        self.dll.AUTDTransitionModeNone.restype = TransitionModeWrap

        self.dll.AUTDGainBessel.argtypes = [Point3, Vector3, Angle, BesselOption]
        self.dll.AUTDGainBessel.restype = GainPtr

        self.dll.AUTDGainBesselIsDefault.argtypes = [BesselOption]
        self.dll.AUTDGainBesselIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainCache.argtypes = [GainPtr]
        self.dll.AUTDGainCache.restype = GainCachePtr

        self.dll.AUTDGainCacheClone.argtypes = [GainCachePtr]
        self.dll.AUTDGainCacheClone.restype = GainPtr

        self.dll.AUTDGainCacheFree.argtypes = [GainCachePtr]
        self.dll.AUTDGainCacheFree.restype = None

        self.dll.AUTDGainCustom.argtypes = [ctypes.c_void_p, ctypes.c_void_p, GeometryPtr]
        self.dll.AUTDGainCustom.restype = GainPtr

        self.dll.AUTDGainFocus.argtypes = [Point3, FocusOption]
        self.dll.AUTDGainFocus.restype = GainPtr

        self.dll.AUTDGainFocusIsDefault.argtypes = [FocusOption]
        self.dll.AUTDGainFocusIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainGroupCreateMap.argtypes = [ctypes.POINTER(ctypes.c_uint16), ctypes.c_uint16]
        self.dll.AUTDGainGroupCreateMap.restype = GroupGainMapPtr

        self.dll.AUTDGainGroupMapSet.argtypes = [GroupGainMapPtr, ctypes.c_uint16, ctypes.POINTER(ctypes.c_int32)]
        self.dll.AUTDGainGroupMapSet.restype = GroupGainMapPtr

        self.dll.AUTDGainGroup.argtypes = [GroupGainMapPtr, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(GainPtr), ctypes.c_uint32]
        self.dll.AUTDGainGroup.restype = GainPtr

        self.dll.AUTDGainIntoDatagramWithSegment.argtypes = [GainPtr, ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDGainIntoDatagramWithSegment.restype = DatagramPtr

        self.dll.AUTDGainIntoDatagram.argtypes = [GainPtr]
        self.dll.AUTDGainIntoDatagram.restype = DatagramPtr

        self.dll.AUTDGainNull.argtypes = []
        self.dll.AUTDGainNull.restype = GainPtr

        self.dll.AUTDGainPlane.argtypes = [Vector3, PlaneOption]
        self.dll.AUTDGainPlane.restype = GainPtr

        self.dll.AUTDGainPlanelIsDefault.argtypes = [PlaneOption]
        self.dll.AUTDGainPlanelIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainUniform.argtypes = [EmitIntensity, Phase]
        self.dll.AUTDGainUniform.restype = GainPtr

        self.dll.AUTDDevice.argtypes = [GeometryPtr, ctypes.c_uint16]
        self.dll.AUTDDevice.restype = DevicePtr

        self.dll.AUTDDeviceNumTransducers.argtypes = [DevicePtr]
        self.dll.AUTDDeviceNumTransducers.restype = ctypes.c_uint32

        self.dll.AUTDDeviceGetSoundSpeed.argtypes = [DevicePtr]
        self.dll.AUTDDeviceGetSoundSpeed.restype = ctypes.c_float

        self.dll.AUTDDeviceSetSoundSpeed.argtypes = [GeometryPtr, ctypes.c_uint16, ctypes.c_float]
        self.dll.AUTDDeviceSetSoundSpeed.restype = None

        self.dll.AUTDDeviceSetSoundSpeedFromTemp.argtypes = [
            GeometryPtr,
            ctypes.c_uint16,
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
        ]
        self.dll.AUTDDeviceSetSoundSpeedFromTemp.restype = None

        self.dll.AUTDDeviceCenter.argtypes = [DevicePtr]
        self.dll.AUTDDeviceCenter.restype = Point3

        self.dll.AUTDDeviceTranslate.argtypes = [GeometryPtr, ctypes.c_uint16, Vector3]
        self.dll.AUTDDeviceTranslate.restype = None

        self.dll.AUTDDeviceRotate.argtypes = [GeometryPtr, ctypes.c_uint16, Quaternion]
        self.dll.AUTDDeviceRotate.restype = None

        self.dll.AUTDDeviceAffine.argtypes = [GeometryPtr, ctypes.c_uint16, Vector3, Quaternion]
        self.dll.AUTDDeviceAffine.restype = None

        self.dll.AUTDDeviceEnableSet.argtypes = [GeometryPtr, ctypes.c_uint16, ctypes.c_bool]
        self.dll.AUTDDeviceEnableSet.restype = None

        self.dll.AUTDDeviceEnableGet.argtypes = [DevicePtr]
        self.dll.AUTDDeviceEnableGet.restype = ctypes.c_bool

        self.dll.AUTDDeviceWavelength.argtypes = [DevicePtr]
        self.dll.AUTDDeviceWavelength.restype = ctypes.c_float

        self.dll.AUTDDeviceWavenumber.argtypes = [DevicePtr]
        self.dll.AUTDDeviceWavenumber.restype = ctypes.c_float

        self.dll.AUTDDeviceRotation.argtypes = [DevicePtr]
        self.dll.AUTDDeviceRotation.restype = Quaternion

        self.dll.AUTDDeviceDirectionX.argtypes = [DevicePtr]
        self.dll.AUTDDeviceDirectionX.restype = Vector3

        self.dll.AUTDDeviceDirectionY.argtypes = [DevicePtr]
        self.dll.AUTDDeviceDirectionY.restype = Vector3

        self.dll.AUTDDeviceDirectionAxial.argtypes = [DevicePtr]
        self.dll.AUTDDeviceDirectionAxial.restype = Vector3

        self.dll.AUTDGeometry.argtypes = [ControllerPtr]
        self.dll.AUTDGeometry.restype = GeometryPtr

        self.dll.AUTDGeometryNumDevices.argtypes = [GeometryPtr]
        self.dll.AUTDGeometryNumDevices.restype = ctypes.c_uint32

        self.dll.AUTDGeometryNumTransducers.argtypes = [GeometryPtr]
        self.dll.AUTDGeometryNumTransducers.restype = ctypes.c_uint32

        self.dll.AUTDGeometrCenter.argtypes = [GeometryPtr]
        self.dll.AUTDGeometrCenter.restype = Point3

        self.dll.AUTDRotationFromEulerXYZ.argtypes = [ctypes.c_float, ctypes.c_float, ctypes.c_float]
        self.dll.AUTDRotationFromEulerXYZ.restype = Quaternion

        self.dll.AUTDRotationFromEulerZYZ.argtypes = [ctypes.c_float, ctypes.c_float, ctypes.c_float]
        self.dll.AUTDRotationFromEulerZYZ.restype = Quaternion

        self.dll.AUTDTransducer.argtypes = [DevicePtr, ctypes.c_uint8]
        self.dll.AUTDTransducer.restype = TransducerPtr

        self.dll.AUTDTransducerPosition.argtypes = [TransducerPtr]
        self.dll.AUTDTransducerPosition.restype = Point3

        self.dll.AUTDTracingInit.argtypes = []
        self.dll.AUTDTracingInit.restype = None

        self.dll.AUTDTracingInitWithFile.argtypes = [ctypes.c_char_p]
        self.dll.AUTDTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDLinkAudit.argtypes = []
        self.dll.AUTDLinkAudit.restype = LinkPtr

        self.dll.AUTDLinkAuditIsOpen.argtypes = [LinkPtr]
        self.dll.AUTDLinkAuditIsOpen.restype = ctypes.c_bool

        self.dll.AUTDLinkAuditBreakDown.argtypes = [LinkPtr]
        self.dll.AUTDLinkAuditBreakDown.restype = None

        self.dll.AUTDLinkAuditRepair.argtypes = [LinkPtr]
        self.dll.AUTDLinkAuditRepair.restype = None

        self.dll.AUTDLinkAuditCpuNumTransducers.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditCpuNumTransducers.restype = ctypes.c_uint32

        self.dll.AUTDLinkAuditFpgaAssertThermalSensor.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaAssertThermalSensor.restype = None

        self.dll.AUTDLinkAuditFpgaDeassertThermalSensor.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaDeassertThermalSensor.restype = None

        self.dll.AUTDLinkAuditFpgaIsForceFan.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaIsForceFan.restype = ctypes.c_bool

        self.dll.AUTDLinkAuditFpgaCurrentStmSegment.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaCurrentStmSegment.restype = Segment

        self.dll.AUTDLinkAuditFpgaCurrentModSegment.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaCurrentModSegment.restype = Segment

        self.dll.AUTDLinkAuditFpgaIsStmGainMode.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaIsStmGainMode.restype = ctypes.c_bool

        self.dll.AUTDLinkAuditCpuSilencerStrictMode.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditCpuSilencerStrictMode.restype = ctypes.c_bool

        self.dll.AUTDLinkAuditFpgaSilencerUpdateRateIntensity.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerUpdateRateIntensity.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaSilencerUpdateRatePhase.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerUpdateRatePhase.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsIntensity.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsIntensity.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsPhase.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsPhase.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaSilencerFixedCompletionStepsMode.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerFixedCompletionStepsMode.restype = ctypes.c_bool

        self.dll.AUTDLinkAuditFpgaSilencerTarget.argtypes = [LinkPtr, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSilencerTarget.restype = SilencerTarget

        self.dll.AUTDLinkAuditFpgaDebugTypes.argtypes = [LinkPtr, ctypes.c_uint16, ctypes.POINTER(ctypes.c_uint8)]
        self.dll.AUTDLinkAuditFpgaDebugTypes.restype = None

        self.dll.AUTDLinkAuditFpgaDebugValues.argtypes = [LinkPtr, ctypes.c_uint16, ctypes.POINTER(ctypes.c_uint64)]
        self.dll.AUTDLinkAuditFpgaDebugValues.restype = None

        self.dll.AUTDLinkAuditFpgaStmFreqDivision.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaStmFreqDivision.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaStmCycle.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaStmCycle.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaSoundSpeed.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaSoundSpeed.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaStmLoopBehavior.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaStmLoopBehavior.restype = LoopBehavior

        self.dll.AUTDLinkAuditFpgaModulationFreqDivision.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaModulationFreqDivision.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaModulationCycle.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaModulationCycle.restype = ctypes.c_uint16

        self.dll.AUTDLinkAuditFpgaModulationBuffer.argtypes = [
            LinkPtr,
            ctypes.c_uint8,
            ctypes.c_uint16,
            ctypes.POINTER(ctypes.c_uint8),
            ctypes.c_uint32,
        ]
        self.dll.AUTDLinkAuditFpgaModulationBuffer.restype = None

        self.dll.AUTDLinkAuditFpgaModulationLoopBehavior.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16]
        self.dll.AUTDLinkAuditFpgaModulationLoopBehavior.restype = LoopBehavior

        self.dll.AUTDLinkAuditFpgaDrivesAt.argtypes = [LinkPtr, ctypes.c_uint8, ctypes.c_uint16, ctypes.c_uint16, ctypes.POINTER(Drive)]
        self.dll.AUTDLinkAuditFpgaDrivesAt.restype = None

        self.dll.AUTDLinkAuditFpgaPulseWidthEncoderTable.argtypes = [LinkPtr, ctypes.c_uint16, ctypes.POINTER(ctypes.c_uint8)]
        self.dll.AUTDLinkAuditFpgaPulseWidthEncoderTable.restype = None

        self.dll.AUTDLinkGet.argtypes = [ControllerPtr]
        self.dll.AUTDLinkGet.restype = LinkPtr

        self.dll.AUTDLinkNop.argtypes = []
        self.dll.AUTDLinkNop.restype = LinkPtr

        self.dll.AUTDModulationCache.argtypes = [ModulationPtr]
        self.dll.AUTDModulationCache.restype = ModulationCachePtr

        self.dll.AUTDModulationCacheClone.argtypes = [ModulationCachePtr]
        self.dll.AUTDModulationCacheClone.restype = ModulationPtr

        self.dll.AUTDModulationCacheFree.argtypes = [ModulationCachePtr]
        self.dll.AUTDModulationCacheFree.restype = None

        self.dll.AUTDModulationCustom.argtypes = [ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16, SamplingConfigWrap]
        self.dll.AUTDModulationCustom.restype = ModulationPtr

        self.dll.AUTDModulationWithFir.argtypes = [ModulationPtr, ctypes.POINTER(ctypes.c_float), ctypes.c_uint32]
        self.dll.AUTDModulationWithFir.restype = ModulationPtr

        self.dll.AUTDModulationFourierExact.argtypes = [ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(SineOption), ctypes.c_uint32, FourierOption]
        self.dll.AUTDModulationFourierExact.restype = ModulationPtr

        self.dll.AUTDModulationFourierExactFloat.argtypes = [
            ctypes.POINTER(ctypes.c_float),
            ctypes.POINTER(SineOption),
            ctypes.c_uint32,
            FourierOption,
        ]
        self.dll.AUTDModulationFourierExactFloat.restype = ModulationPtr

        self.dll.AUTDModulationFourierNearest.argtypes = [ctypes.POINTER(ctypes.c_float), ctypes.POINTER(SineOption), ctypes.c_uint32, FourierOption]
        self.dll.AUTDModulationFourierNearest.restype = ModulationPtr

        self.dll.AUTDModulationSamplingConfig.argtypes = [ModulationPtr]
        self.dll.AUTDModulationSamplingConfig.restype = SamplingConfigWrap

        self.dll.AUTDModulationIntoDatagramWithSegment.argtypes = [ModulationPtr, ctypes.c_uint8, TransitionModeWrap]
        self.dll.AUTDModulationIntoDatagramWithSegment.restype = DatagramPtr

        self.dll.AUTDModulationIntoDatagramWithLoopBehavior.argtypes = [ModulationPtr, ctypes.c_uint8, TransitionModeWrap, LoopBehavior]
        self.dll.AUTDModulationIntoDatagramWithLoopBehavior.restype = DatagramPtr

        self.dll.AUTDModulationIntoDatagram.argtypes = [ModulationPtr]
        self.dll.AUTDModulationIntoDatagram.restype = DatagramPtr

        self.dll.AUTDModulationWithRadiationPressure.argtypes = [ModulationPtr]
        self.dll.AUTDModulationWithRadiationPressure.restype = ModulationPtr

        self.dll.AUTDModulationSineExact.argtypes = [ctypes.c_uint32, SineOption]
        self.dll.AUTDModulationSineExact.restype = ModulationPtr

        self.dll.AUTDModulationSineExactFloat.argtypes = [ctypes.c_float, SineOption]
        self.dll.AUTDModulationSineExactFloat.restype = ModulationPtr

        self.dll.AUTDModulationSineNearest.argtypes = [ctypes.c_float, SineOption]
        self.dll.AUTDModulationSineNearest.restype = ModulationPtr

        self.dll.AUTDModulationSineIsDefault.argtypes = [SineOption]
        self.dll.AUTDModulationSineIsDefault.restype = ctypes.c_bool

        self.dll.AUTDModulationSquareExact.argtypes = [ctypes.c_uint32, SquareOption]
        self.dll.AUTDModulationSquareExact.restype = ModulationPtr

        self.dll.AUTDModulationSquareExactFloat.argtypes = [ctypes.c_float, SquareOption]
        self.dll.AUTDModulationSquareExactFloat.restype = ModulationPtr

        self.dll.AUTDModulationSquareNearest.argtypes = [ctypes.c_float, SquareOption]
        self.dll.AUTDModulationSquareNearest.restype = ModulationPtr

        self.dll.AUTDModulationSquareIsDefault.argtypes = [SquareOption]
        self.dll.AUTDModulationSquareIsDefault.restype = ctypes.c_bool

        self.dll.AUTDModulationStatic.argtypes = [ctypes.c_uint8]
        self.dll.AUTDModulationStatic.restype = ModulationPtr

        self.dll.AUTDModulationStaticIsDefault.argtypes = [ctypes.c_uint8]
        self.dll.AUTDModulationStaticIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGetErr.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
        self.dll.AUTDGetErr.restype = None

    def controller_group(
        self,
        sender: SenderPtr,
        f: ctypes.c_void_p,
        context: ctypes.c_void_p,
        geometry: GeometryPtr,
        keys: ctypes.Array[ctypes.c_int32],
        d: ctypes.Array[DatagramPtr],
        n: int,
    ) -> ResultStatus:
        return self.dll.AUTDControllerGroup(sender, f, context, geometry, keys, d, n)

    def controller_open(
        self,
        pos: ctypes.Array[Point3],
        rot: ctypes.Array[Quaternion],
        len_: int,
        link: LinkPtr,
        option: SenderOption,
    ) -> ResultController:
        return self.dll.AUTDControllerOpen(pos, rot, len_, link, option)

    def controller_close(self, cnt: ControllerPtr) -> ResultStatus:
        return self.dll.AUTDControllerClose(cnt)

    def controller_fpga_state(self, cnt: ControllerPtr) -> ResultFPGAStateList:
        return self.dll.AUTDControllerFPGAState(cnt)

    def controller_fpga_state_get(self, p: FPGAStateListPtr, idx: int) -> ctypes.c_int16:
        return self.dll.AUTDControllerFPGAStateGet(p, idx)

    def controller_fpga_state_delete(self, p: FPGAStateListPtr) -> None:
        return self.dll.AUTDControllerFPGAStateDelete(p)

    def controller_firmware_version_list_pointer(self, cnt: ControllerPtr) -> ResultFirmwareVersionList:
        return self.dll.AUTDControllerFirmwareVersionListPointer(cnt)

    def controller_firmware_version_get(self, p_info_list: FirmwareVersionListPtr, idx: int, info: bytes) -> None:
        return self.dll.AUTDControllerFirmwareVersionGet(p_info_list, idx, info)

    def controller_firmware_version_list_pointer_delete(self, p_info_list: FirmwareVersionListPtr) -> None:
        return self.dll.AUTDControllerFirmwareVersionListPointerDelete(p_info_list)

    def firmware_latest(self, latest: bytes) -> None:
        return self.dll.AUTDFirmwareLatest(latest)

    def sender(self, cnt: ControllerPtr, option: SenderOption) -> SenderPtr:
        return self.dll.AUTDSender(cnt, option)

    def sender_send(self, sender: SenderPtr, d: DatagramPtr) -> ResultStatus:
        return self.dll.AUTDSenderSend(sender, d)

    def spin_sleep_default_accuracy(self) -> ctypes.c_uint32:
        return self.dll.AUTDSpinSleepDefaultAccuracy()

    def sender_option_is_default(self, option: SenderOption) -> ctypes.c_bool:
        return self.dll.AUTDSenderOptionIsDefault(option)

    def datagram_clear(self) -> DatagramPtr:
        return self.dll.AUTDDatagramClear()

    def datagram_debug_settings(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramDebugSettings(f, context, geometry)

    def datagram_force_fan(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramForceFan(f, context, geometry)

    def datagram_tuple(self, d1: DatagramPtr, d2: DatagramPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramTuple(d1, d2)

    def datagram_phase_corr(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramPhaseCorr(f, context, geometry)

    def datagram_pulse_width_encoder(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramPulseWidthEncoder(f, context, geometry)

    def datagram_pulse_width_encoder_default(self) -> DatagramPtr:
        return self.dll.AUTDDatagramPulseWidthEncoderDefault()

    def datagram_reads_fpga_state(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> DatagramPtr:
        return self.dll.AUTDDatagramReadsFPGAState(f, context, geometry)

    def datagram_swap_segment_modulation(self, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDDatagramSwapSegmentModulation(segment, transition_mode)

    def datagram_swap_segment_foci_stm(self, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDDatagramSwapSegmentFociSTM(segment, transition_mode)

    def datagram_swap_segment_gain_stm(self, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDDatagramSwapSegmentGainSTM(segment, transition_mode)

    def datagram_swap_segment_gain(self, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDDatagramSwapSegmentGain(segment, transition_mode)

    def datagram_silencer_from_update_rate(self, config: FixedUpdateRate, target: SilencerTarget) -> DatagramPtr:
        return self.dll.AUTDDatagramSilencerFromUpdateRate(config, target)

    def datagram_silencer_from_completion_steps(self, config: FixedCompletionSteps, target: SilencerTarget) -> DatagramPtr:
        return self.dll.AUTDDatagramSilencerFromCompletionSteps(config, target)

    def datagram_silencer_from_completion_time(self, config: FixedCompletionTime, target: SilencerTarget) -> DatagramPtr:
        return self.dll.AUTDDatagramSilencerFromCompletionTime(config, target)

    def datagram_silencer_fixed_completion_steps_is_default(self, config: FixedCompletionSteps, target: SilencerTarget) -> ctypes.c_bool:
        return self.dll.AUTDDatagramSilencerFixedCompletionStepsIsDefault(config, target)

    def stm_foci(self, config: SamplingConfigWrap, points: ctypes.c_void_p, size: int, n: int) -> FociSTMPtr:
        return self.dll.AUTDSTMFoci(config, points, size, n)

    def stm_foci_into_datagram_with_segment(self, stm: FociSTMPtr, n: int, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDSTMFociIntoDatagramWithSegment(stm, n, segment, transition_mode)

    def stm_foci_into_datagram_with_loop_behavior(
        self,
        stm: FociSTMPtr,
        n: int,
        segment: Segment,
        transition_mode: TransitionModeWrap,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return self.dll.AUTDSTMFociIntoDatagramWithLoopBehavior(stm, n, segment, transition_mode, loop_behavior)

    def stm_foci_into_datagram(self, stm: FociSTMPtr, n: int) -> DatagramPtr:
        return self.dll.AUTDSTMFociIntoDatagram(stm, n)

    def stm_gain(self, config: SamplingConfigWrap, gains: ctypes.Array[GainPtr], size: int, option: GainSTMOption) -> GainSTMPtr:
        return self.dll.AUTDSTMGain(config, gains, size, option)

    def stm_gain_into_datagram_with_segment(self, stm: GainSTMPtr, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDSTMGainIntoDatagramWithSegment(stm, segment, transition_mode)

    def stm_gain_into_datagram_with_loop_behavior(
        self,
        stm: GainSTMPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return self.dll.AUTDSTMGainIntoDatagramWithLoopBehavior(stm, segment, transition_mode, loop_behavior)

    def stm_gain_into_datagram(self, stm: GainSTMPtr) -> DatagramPtr:
        return self.dll.AUTDSTMGainIntoDatagram(stm)

    def stm_config_from_freq(self, f: float, n: int) -> ResultSamplingConfig:
        return self.dll.AUTDSTMConfigFromFreq(f, n)

    def stm_config_from_period(self, p: Duration, n: int) -> ResultSamplingConfig:
        return self.dll.AUTDSTMConfigFromPeriod(p, n)

    def stm_config_from_freq_nearest(self, f: float, n: int) -> SamplingConfigWrap:
        return self.dll.AUTDSTMConfigFromFreqNearest(f, n)

    def stm_config_from_period_nearest(self, p: Duration, n: int) -> SamplingConfigWrap:
        return self.dll.AUTDSTMConfigFromPeriodNearest(p, n)

    def datagram_synchronize(self) -> DatagramPtr:
        return self.dll.AUTDDatagramSynchronize()

    def dc_sys_time_now(self) -> DcSysTime:
        return self.dll.AUTDDcSysTimeNow()

    def debug_type_none(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeNone()

    def debug_type_base_signal(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeBaseSignal()

    def debug_type_thermo(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeThermo()

    def debug_type_force_fan(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeForceFan()

    def debug_type_sync(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeSync()

    def debug_type_mod_segment(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeModSegment()

    def debug_type_mod_idx(self, value: int) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeModIdx(value)

    def debug_type_stm_segment(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeStmSegment()

    def debug_type_stm_idx(self, value: int) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeStmIdx(value)

    def debug_type_is_stm_mode(self) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeIsStmMode()

    def debug_type_pwm_out(self, value: TransducerPtr) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypePwmOut(value)

    def debug_type_direct(self, value: bool) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeDirect(value)

    def debug_type_sys_time_eq(self, sys_time: DcSysTime) -> DebugTypeWrap:
        return self.dll.AUTDDebugTypeSysTimeEq(sys_time)

    def loop_behavior_infinite(self) -> LoopBehavior:
        return self.dll.AUTDLoopBehaviorInfinite()

    def loop_behavior_finite(self, v: int) -> LoopBehavior:
        return self.dll.AUTDLoopBehaviorFinite(v)

    def phase_from_rad(self, value: float) -> ctypes.c_uint8:
        return self.dll.AUTDPhaseFromRad(value)

    def phase_to_rad(self, value: Phase) -> ctypes.c_float:
        return self.dll.AUTDPhaseToRad(value)

    def sampling_config_from_division(self, div: int) -> ResultSamplingConfig:
        return self.dll.AUTDSamplingConfigFromDivision(div)

    def sampling_config_from_freq(self, f: float) -> SamplingConfigWrap:
        return self.dll.AUTDSamplingConfigFromFreq(f)

    def sampling_config_from_period(self, p: Duration) -> SamplingConfigWrap:
        return self.dll.AUTDSamplingConfigFromPeriod(p)

    def sampling_config_into_nearest(self, config: SamplingConfigWrap) -> SamplingConfigWrap:
        return self.dll.AUTDSamplingConfigIntoNearest(config)

    def sampling_config_division(self, c: SamplingConfigWrap) -> ResultU16:
        return self.dll.AUTDSamplingConfigDivision(c)

    def sampling_config_freq(self, c: SamplingConfigWrap) -> ResultF32:
        return self.dll.AUTDSamplingConfigFreq(c)

    def sampling_config_period(self, c: SamplingConfigWrap) -> ResultDuration:
        return self.dll.AUTDSamplingConfigPeriod(c)

    def sampling_config_eq(self, a: SamplingConfigWrap, b: SamplingConfigWrap) -> ctypes.c_bool:
        return self.dll.AUTDSamplingConfigEq(a, b)

    def transition_mode_sync_idx(self) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeSyncIdx()

    def transition_mode_sys_time(self, sys_time: DcSysTime) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeSysTime(sys_time)

    def transition_mode_gpio(self, gpio: GPIOIn) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeGPIO(gpio)

    def transition_mode_ext(self) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeExt()

    def transition_mode_immediate(self) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeImmediate()

    def transition_mode_none(self) -> TransitionModeWrap:
        return self.dll.AUTDTransitionModeNone()

    def gain_bessel(self, pos: Point3, dir_: Vector3, theta: Angle, option: BesselOption) -> GainPtr:
        return self.dll.AUTDGainBessel(pos, dir_, theta, option)

    def gain_bessel_is_default(self, option: BesselOption) -> ctypes.c_bool:
        return self.dll.AUTDGainBesselIsDefault(option)

    def gain_cache(self, g: GainPtr) -> GainCachePtr:
        return self.dll.AUTDGainCache(g)

    def gain_cache_clone(self, g: GainCachePtr) -> GainPtr:
        return self.dll.AUTDGainCacheClone(g)

    def gain_cache_free(self, g: GainCachePtr) -> None:
        return self.dll.AUTDGainCacheFree(g)

    def gain_custom(self, f: ctypes.c_void_p, context: ctypes.c_void_p, geometry: GeometryPtr) -> GainPtr:
        return self.dll.AUTDGainCustom(f, context, geometry)

    def gain_focus(self, pos: Point3, option: FocusOption) -> GainPtr:
        return self.dll.AUTDGainFocus(pos, option)

    def gain_focus_is_default(self, option: FocusOption) -> ctypes.c_bool:
        return self.dll.AUTDGainFocusIsDefault(option)

    def gain_group_create_map(self, device_indices_ptr: ctypes.Array[ctypes.c_uint16], num_devices: int) -> GroupGainMapPtr:
        return self.dll.AUTDGainGroupCreateMap(device_indices_ptr, num_devices)

    def gain_group_map_set(self, map_: GroupGainMapPtr, dev_idx: int, map_data: ctypes.Array[ctypes.c_int32]) -> GroupGainMapPtr:
        return self.dll.AUTDGainGroupMapSet(map_, dev_idx, map_data)

    def gain_group(self, map_: GroupGainMapPtr, keys_ptr: ctypes.Array[ctypes.c_int32], values_ptr: ctypes.Array[GainPtr], kv_len: int) -> GainPtr:
        return self.dll.AUTDGainGroup(map_, keys_ptr, values_ptr, kv_len)

    def gain_into_datagram_with_segment(self, gain: GainPtr, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDGainIntoDatagramWithSegment(gain, segment, transition_mode)

    def gain_into_datagram(self, gain: GainPtr) -> DatagramPtr:
        return self.dll.AUTDGainIntoDatagram(gain)

    def gain_null(self) -> GainPtr:
        return self.dll.AUTDGainNull()

    def gain_plane(self, n: Vector3, option: PlaneOption) -> GainPtr:
        return self.dll.AUTDGainPlane(n, option)

    def gain_planel_is_default(self, option: PlaneOption) -> ctypes.c_bool:
        return self.dll.AUTDGainPlanelIsDefault(option)

    def gain_uniform(self, intensity: EmitIntensity, phase: Phase) -> GainPtr:
        return self.dll.AUTDGainUniform(intensity, phase)

    def device(self, geo: GeometryPtr, dev_idx: int) -> DevicePtr:
        return self.dll.AUTDDevice(geo, dev_idx)

    def device_num_transducers(self, dev: DevicePtr) -> ctypes.c_uint32:
        return self.dll.AUTDDeviceNumTransducers(dev)

    def device_get_sound_speed(self, dev: DevicePtr) -> ctypes.c_float:
        return self.dll.AUTDDeviceGetSoundSpeed(dev)

    def device_set_sound_speed(self, geo: GeometryPtr, dev_idx: int, value: float) -> None:
        return self.dll.AUTDDeviceSetSoundSpeed(geo, dev_idx, value)

    def device_set_sound_speed_from_temp(self, geo: GeometryPtr, dev_idx: int, temp: float, k: float, r: float, m: float) -> None:
        return self.dll.AUTDDeviceSetSoundSpeedFromTemp(geo, dev_idx, temp, k, r, m)

    def device_center(self, dev: DevicePtr) -> Point3:
        return self.dll.AUTDDeviceCenter(dev)

    def device_translate(self, geo: GeometryPtr, dev_idx: int, t: Vector3) -> None:
        return self.dll.AUTDDeviceTranslate(geo, dev_idx, t)

    def device_rotate(self, geo: GeometryPtr, dev_idx: int, r: Quaternion) -> None:
        return self.dll.AUTDDeviceRotate(geo, dev_idx, r)

    def device_affine(self, geo: GeometryPtr, dev_idx: int, t: Vector3, r: Quaternion) -> None:
        return self.dll.AUTDDeviceAffine(geo, dev_idx, t, r)

    def device_enable_set(self, geo: GeometryPtr, dev_idx: int, value: bool) -> None:
        return self.dll.AUTDDeviceEnableSet(geo, dev_idx, value)

    def device_enable_get(self, dev: DevicePtr) -> ctypes.c_bool:
        return self.dll.AUTDDeviceEnableGet(dev)

    def device_wavelength(self, dev: DevicePtr) -> ctypes.c_float:
        return self.dll.AUTDDeviceWavelength(dev)

    def device_wavenumber(self, dev: DevicePtr) -> ctypes.c_float:
        return self.dll.AUTDDeviceWavenumber(dev)

    def device_rotation(self, dev: DevicePtr) -> Quaternion:
        return self.dll.AUTDDeviceRotation(dev)

    def device_direction_x(self, dev: DevicePtr) -> Vector3:
        return self.dll.AUTDDeviceDirectionX(dev)

    def device_direction_y(self, dev: DevicePtr) -> Vector3:
        return self.dll.AUTDDeviceDirectionY(dev)

    def device_direction_axial(self, dev: DevicePtr) -> Vector3:
        return self.dll.AUTDDeviceDirectionAxial(dev)

    def geometry(self, cnt: ControllerPtr) -> GeometryPtr:
        return self.dll.AUTDGeometry(cnt)

    def geometry_num_devices(self, geo: GeometryPtr) -> ctypes.c_uint32:
        return self.dll.AUTDGeometryNumDevices(geo)

    def geometry_num_transducers(self, geo: GeometryPtr) -> ctypes.c_uint32:
        return self.dll.AUTDGeometryNumTransducers(geo)

    def geometr_center(self, geo: GeometryPtr) -> Point3:
        return self.dll.AUTDGeometrCenter(geo)

    def rotation_from_euler_xyz(self, x: float, y: float, z: float) -> Quaternion:
        return self.dll.AUTDRotationFromEulerXYZ(x, y, z)

    def rotation_from_euler_zyz(self, z1: float, y: float, z2: float) -> Quaternion:
        return self.dll.AUTDRotationFromEulerZYZ(z1, y, z2)

    def transducer(self, dev: DevicePtr, idx: int) -> TransducerPtr:
        return self.dll.AUTDTransducer(dev, idx)

    def transducer_position(self, tr: TransducerPtr) -> Point3:
        return self.dll.AUTDTransducerPosition(tr)

    def tracing_init(self) -> None:
        return self.dll.AUTDTracingInit()

    def tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDTracingInitWithFile(path)

    def link_audit(self) -> LinkPtr:
        return self.dll.AUTDLinkAudit()

    def link_audit_is_open(self, audit: LinkPtr) -> ctypes.c_bool:
        return self.dll.AUTDLinkAuditIsOpen(audit)

    def link_audit_break_down(self, audit: LinkPtr) -> None:
        return self.dll.AUTDLinkAuditBreakDown(audit)

    def link_audit_repair(self, audit: LinkPtr) -> None:
        return self.dll.AUTDLinkAuditRepair(audit)

    def link_audit_cpu_num_transducers(self, audit: LinkPtr, idx: int) -> ctypes.c_uint32:
        return self.dll.AUTDLinkAuditCpuNumTransducers(audit, idx)

    def link_audit_fpga_assert_thermal_sensor(self, audit: LinkPtr, idx: int) -> None:
        return self.dll.AUTDLinkAuditFpgaAssertThermalSensor(audit, idx)

    def link_audit_fpga_deassert_thermal_sensor(self, audit: LinkPtr, idx: int) -> None:
        return self.dll.AUTDLinkAuditFpgaDeassertThermalSensor(audit, idx)

    def link_audit_fpga_is_force_fan(self, audit: LinkPtr, idx: int) -> ctypes.c_bool:
        return self.dll.AUTDLinkAuditFpgaIsForceFan(audit, idx)

    def link_audit_fpga_current_stm_segment(self, audit: LinkPtr, idx: int) -> Segment:
        return self.dll.AUTDLinkAuditFpgaCurrentStmSegment(audit, idx)

    def link_audit_fpga_current_mod_segment(self, audit: LinkPtr, idx: int) -> Segment:
        return self.dll.AUTDLinkAuditFpgaCurrentModSegment(audit, idx)

    def link_audit_fpga_is_stm_gain_mode(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_bool:
        return self.dll.AUTDLinkAuditFpgaIsStmGainMode(audit, segment, idx)

    def link_audit_cpu_silencer_strict_mode(self, audit: LinkPtr, idx: int) -> ctypes.c_bool:
        return self.dll.AUTDLinkAuditCpuSilencerStrictMode(audit, idx)

    def link_audit_fpga_silencer_update_rate_intensity(self, audit: LinkPtr, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaSilencerUpdateRateIntensity(audit, idx)

    def link_audit_fpga_silencer_update_rate_phase(self, audit: LinkPtr, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaSilencerUpdateRatePhase(audit, idx)

    def link_audit_fpga_silencer_completion_steps_intensity(self, audit: LinkPtr, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsIntensity(audit, idx)

    def link_audit_fpga_silencer_completion_steps_phase(self, audit: LinkPtr, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaSilencerCompletionStepsPhase(audit, idx)

    def link_audit_fpga_silencer_fixed_completion_steps_mode(self, audit: LinkPtr, idx: int) -> ctypes.c_bool:
        return self.dll.AUTDLinkAuditFpgaSilencerFixedCompletionStepsMode(audit, idx)

    def link_audit_fpga_silencer_target(self, audit: LinkPtr, idx: int) -> SilencerTarget:
        return self.dll.AUTDLinkAuditFpgaSilencerTarget(audit, idx)

    def link_audit_fpga_debug_types(self, audit: LinkPtr, idx: int, ty: ctypes.Array[ctypes.c_uint8]) -> None:
        return self.dll.AUTDLinkAuditFpgaDebugTypes(audit, idx, ty)

    def link_audit_fpga_debug_values(self, audit: LinkPtr, idx: int, value: ctypes.Array[ctypes.c_uint64]) -> None:
        return self.dll.AUTDLinkAuditFpgaDebugValues(audit, idx, value)

    def link_audit_fpga_stm_freq_division(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaStmFreqDivision(audit, segment, idx)

    def link_audit_fpga_stm_cycle(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaStmCycle(audit, segment, idx)

    def link_audit_fpga_sound_speed(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaSoundSpeed(audit, segment, idx)

    def link_audit_fpga_stm_loop_behavior(self, audit: LinkPtr, segment: Segment, idx: int) -> LoopBehavior:
        return self.dll.AUTDLinkAuditFpgaStmLoopBehavior(audit, segment, idx)

    def link_audit_fpga_modulation_freq_division(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaModulationFreqDivision(audit, segment, idx)

    def link_audit_fpga_modulation_cycle(self, audit: LinkPtr, segment: Segment, idx: int) -> ctypes.c_uint16:
        return self.dll.AUTDLinkAuditFpgaModulationCycle(audit, segment, idx)

    def link_audit_fpga_modulation_buffer(self, audit: LinkPtr, segment: Segment, idx: int, data: ctypes.Array[ctypes.c_uint8], size: int) -> None:
        return self.dll.AUTDLinkAuditFpgaModulationBuffer(audit, segment, idx, data, size)

    def link_audit_fpga_modulation_loop_behavior(self, audit: LinkPtr, segment: Segment, idx: int) -> LoopBehavior:
        return self.dll.AUTDLinkAuditFpgaModulationLoopBehavior(audit, segment, idx)

    def link_audit_fpga_drives_at(self, audit: LinkPtr, segment: Segment, idx: int, stm_idx: int, drive: ctypes.Array[Drive]) -> None:
        return self.dll.AUTDLinkAuditFpgaDrivesAt(audit, segment, idx, stm_idx, drive)

    def link_audit_fpga_pulse_width_encoder_table(self, audit: LinkPtr, idx: int, dst: ctypes.Array[ctypes.c_uint8]) -> None:
        return self.dll.AUTDLinkAuditFpgaPulseWidthEncoderTable(audit, idx, dst)

    def link_get(self, cnt: ControllerPtr) -> LinkPtr:
        return self.dll.AUTDLinkGet(cnt)

    def link_nop(self) -> LinkPtr:
        return self.dll.AUTDLinkNop()

    def modulation_cache(self, m: ModulationPtr) -> ModulationCachePtr:
        return self.dll.AUTDModulationCache(m)

    def modulation_cache_clone(self, m: ModulationCachePtr) -> ModulationPtr:
        return self.dll.AUTDModulationCacheClone(m)

    def modulation_cache_free(self, m: ModulationCachePtr) -> None:
        return self.dll.AUTDModulationCacheFree(m)

    def modulation_custom(self, ptr: ctypes.Array[ctypes.c_uint8], len_: int, sampling_config: SamplingConfigWrap) -> ModulationPtr:
        return self.dll.AUTDModulationCustom(ptr, len_, sampling_config)

    def modulation_with_fir(self, m: ModulationPtr, coef: ctypes.Array[ctypes.c_float], n_tap: int) -> ModulationPtr:
        return self.dll.AUTDModulationWithFir(m, coef, n_tap)

    def modulation_fourier_exact(
        self,
        sine_freq: ctypes.Array[ctypes.c_uint32],
        sine_clamp: ctypes.Array[SineOption],
        size: int,
        option: FourierOption,
    ) -> ModulationPtr:
        return self.dll.AUTDModulationFourierExact(sine_freq, sine_clamp, size, option)

    def modulation_fourier_exact_float(
        self,
        sine_freq: ctypes.Array[ctypes.c_float],
        sine_clamp: ctypes.Array[SineOption],
        size: int,
        option: FourierOption,
    ) -> ModulationPtr:
        return self.dll.AUTDModulationFourierExactFloat(sine_freq, sine_clamp, size, option)

    def modulation_fourier_nearest(
        self,
        sine_freq: ctypes.Array[ctypes.c_float],
        sine_clamp: ctypes.Array[SineOption],
        size: int,
        option: FourierOption,
    ) -> ModulationPtr:
        return self.dll.AUTDModulationFourierNearest(sine_freq, sine_clamp, size, option)

    def modulation_sampling_config(self, m: ModulationPtr) -> SamplingConfigWrap:
        return self.dll.AUTDModulationSamplingConfig(m)

    def modulation_into_datagram_with_segment(self, m: ModulationPtr, segment: Segment, transition_mode: TransitionModeWrap) -> DatagramPtr:
        return self.dll.AUTDModulationIntoDatagramWithSegment(m, segment, transition_mode)

    def modulation_into_datagram_with_loop_behavior(
        self,
        m: ModulationPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return self.dll.AUTDModulationIntoDatagramWithLoopBehavior(m, segment, transition_mode, loop_behavior)

    def modulation_into_datagram(self, m: ModulationPtr) -> DatagramPtr:
        return self.dll.AUTDModulationIntoDatagram(m)

    def modulation_with_radiation_pressure(self, m: ModulationPtr) -> ModulationPtr:
        return self.dll.AUTDModulationWithRadiationPressure(m)

    def modulation_sine_exact(self, freq: int, option: SineOption) -> ModulationPtr:
        return self.dll.AUTDModulationSineExact(freq, option)

    def modulation_sine_exact_float(self, freq: float, option: SineOption) -> ModulationPtr:
        return self.dll.AUTDModulationSineExactFloat(freq, option)

    def modulation_sine_nearest(self, freq: float, option: SineOption) -> ModulationPtr:
        return self.dll.AUTDModulationSineNearest(freq, option)

    def modulation_sine_is_default(self, option: SineOption) -> ctypes.c_bool:
        return self.dll.AUTDModulationSineIsDefault(option)

    def modulation_square_exact(self, freq: int, option: SquareOption) -> ModulationPtr:
        return self.dll.AUTDModulationSquareExact(freq, option)

    def modulation_square_exact_float(self, freq: float, option: SquareOption) -> ModulationPtr:
        return self.dll.AUTDModulationSquareExactFloat(freq, option)

    def modulation_square_nearest(self, freq: float, option: SquareOption) -> ModulationPtr:
        return self.dll.AUTDModulationSquareNearest(freq, option)

    def modulation_square_is_default(self, option: SquareOption) -> ctypes.c_bool:
        return self.dll.AUTDModulationSquareIsDefault(option)

    def modulation_static(self, intensity: int) -> ModulationPtr:
        return self.dll.AUTDModulationStatic(intensity)

    def modulation_static_is_default(self, intensity: int) -> ctypes.c_bool:
        return self.dll.AUTDModulationStaticIsDefault(intensity)

    def get_err(self, src: ctypes.c_void_p, dst: bytes) -> None:
        return self.dll.AUTDGetErr(src, dst)



from pathlib import Path
from typing import Self

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


class Wav(Modulation):
    path: Path

    def __init__(self: Self, path: Path) -> None:
        super().__init__()
        self.path = path

    def _modulation_ptr(self: Self) -> ModulationPtr:
        path = _to_null_terminated_utf8(str(self.path))
        return _validate_ptr(ModulationAudioFile().modulation_audio_file_wav(path))



from typing import Self

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.utils import _validate_u8
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr


class Static(Modulation):
    intensity: int

    def __init__(self: Self, intensity: int = 0xFF) -> None:
        super().__init__()
        self.intensity = _validate_u8(intensity)

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_static(self.intensity)



from abc import ABCMeta, abstractmethod
from ctypes import Array, c_float
from typing import Self

from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import BackendPtr, GSOption, GSPATOption, LMOption, NaiveOption
from pyautd3.native_methods.structs import Point3


class Backend(metaclass=ABCMeta):
    _ptr: BackendPtr

    def __init__(self: Self, ptr: BackendPtr) -> None:
        self._ptr = ptr

    def _backend_ptr(self: Self) -> BackendPtr:
        return self._ptr

    @abstractmethod
    def _gs(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: GSOption) -> GainPtr:
        pass

    @abstractmethod
    def _gspat(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: GSPATOption) -> GainPtr:
        pass

    @abstractmethod
    def _naive(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: NaiveOption) -> GainPtr:
        pass

    @abstractmethod
    def _lm(self: Self, foci: Array[Point3], amps: Array[c_float], size: int, option: LMOption) -> GainPtr:
        pass



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np

from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.geometry import Geometry
from pyautd3.driver.utils import _validate_nonzero_u16
from pyautd3.gain.holo.amplitude import Amplitude
from pyautd3.gain.holo.constraint import EmissionConstraint
from pyautd3.gain.holo.holo import Holo
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.autd3capi_gain_holo import EmissionConstraintWrap
from pyautd3.native_methods.autd3capi_gain_holo import GreedyOption as GreedyOption_
from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo
from pyautd3.native_methods.structs import Point3


class GreedyOption:
    phase_div: int
    constraint: EmissionConstraintWrap

    def __init__(self: Self, *, phase_div: int = 16, constraint: EmissionConstraintWrap | None = None) -> None:
        self.phase_div = _validate_nonzero_u16(phase_div)
        self.constraint = constraint or EmissionConstraint.Uniform(EmitIntensity.MAX)

    def _inner(self: Self) -> GreedyOption_:
        return GreedyOption_(self.constraint, self.phase_div)


class Greedy(Holo["Greedy"]):
    option: GreedyOption

    def __init__(self: Self, foci: Iterable[tuple[np.ndarray, Amplitude]], option: GreedyOption) -> None:
        super().__init__(foci)
        self.option = option

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        size = len(self.foci)
        points = np.fromiter((np.void(Point3(d)) for d, _ in self.foci), dtype=Point3)  # type: ignore[type-var,call-overload]
        amps = np.fromiter((d.pascal() for _, d in self.foci), dtype=ctypes.c_float)  # type: ignore[type-var,call-overload]
        return GainHolo().gain_holo_greedy_sphere(
            points.ctypes.data_as(ctypes.POINTER(Point3)),  # type: ignore[arg-type]
            amps.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
            size,
            self.option._inner(),
        )



from typing import Generic, Self, TypeVar

T = TypeVar("T", int, float)


class Freq(Generic[T]):
    _freq: T

    def __new__(cls: type["Freq"]) -> "Freq":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["Freq[T]"], freq: T) -> "Freq[T]":
        ins = super().__new__(cls)
        ins._freq = freq
        return ins

    def hz(self: Self) -> T:
        return self._freq

    def __eq__(self: Self, value: object) -> bool:
        return isinstance(value, Freq) and self._freq == value._freq


class _UnitHz:
    def __new__(cls: type["_UnitHz"]) -> "_UnitHz":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["_UnitHz"]) -> "_UnitHz":
        return super().__new__(cls)

    def __rmul__(self: Self, other: T) -> "Freq[T]":
        return Freq.__private_new__(other)


class _UnitkHz:
    def __new__(cls: type["_UnitkHz"]) -> "_UnitkHz":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["_UnitkHz"]) -> "_UnitkHz":
        return super().__new__(cls)

    def __rmul__(self: Self, other: T) -> "Freq[T]":
        return Freq.__private_new__(other * 1000)


Hz: _UnitHz = _UnitHz.__private_new__()
kHz: _UnitkHz = _UnitkHz.__private_new__()  # noqa: N816



from typing import Generic, Self, TypeVar

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr

M = TypeVar("M", bound=Modulation)


class RadiationPressure(Modulation, Generic[M]):
    target: M

    def __init__(self: Self, target: M) -> None:
        self.target = target

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_with_radiation_pressure(self.target._modulation_ptr())



import platform
from typing import Self

from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import SleeperTag, SleeperWrap
from pyautd3.native_methods.autd3capi_driver import SpinStrategyTag as SpinStrategy
from pyautd3.utils import Duration


class StdSleeper:
    timer_resolution: int | None

    def __init__(self: Self, *, timer_resolution: int | None = 1) -> None:
        self.timer_resolution = timer_resolution

    def _inner(self: Self) -> SleeperWrap:
        return SleeperWrap(SleeperTag.Std, self.timer_resolution or 0)


class SpinSleeper:
    native_accuracy: Duration
    spin_strategy: SpinStrategy

    def __init__(
        self: Self,
        native_accuracy: Duration | None = None,
    ) -> None:
        self.native_accuracy = native_accuracy or Duration.from_nanos(int(Base().spin_sleep_default_accuracy()))
        self.spin_strategy = SpinStrategy.SpinLoopHint if platform.system() == "Windows" else SpinStrategy.YieldThread

    def with_spin_strategy(self: Self, spin_strategy: SpinStrategy) -> Self:
        self.spin_strategy = spin_strategy
        return self

    def _inner(self: Self) -> SleeperWrap:
        return SleeperWrap(SleeperTag.Spin, self.native_accuracy.as_nanos(), self.spin_strategy)


class WaitableSleeper:
    def __init__(self: Self) -> None:
        pass

    def _inner(self: Self) -> SleeperWrap:
        if platform.system() != "Windows":  # pragma: no cover
            err = "WaitableSleeper is only supported on Windows"  # pragma: no cover
            raise RuntimeError(err)  # pragma: no cover
        return SleeperWrap(SleeperTag.Waitable, 0)  # pragma: no cover



import ctypes
import enum
import threading
from pathlib import Path

from pyautd3.native_methods.autd3 import EmitIntensity
from pyautd3.native_methods.autd3capi_driver import GainPtr
from pyautd3.native_methods.structs import Point3


class EmissionConstraintTag(enum.IntEnum):
    Normalize = 1
    Uniform = 2
    Multiply = 3
    Clamp = 4

    @classmethod
    def from_param(cls, obj):
        return int(obj)  # pragma: no cover


class EmissionConstraintValue(ctypes.Union):
    _fields_ = [("null", EmitIntensity), ("uniform", EmitIntensity), ("multiply", ctypes.c_float), ("clamp", EmitIntensity * 2)]


class EmissionConstraintWrap(ctypes.Structure):
    _fields_ = [("tag", ctypes.c_uint8), ("value", EmissionConstraintValue)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, EmissionConstraintWrap) and self._fields_ == other._fields_  # pragma: no cover


class LMOption(ctypes.Structure):
    _fields_ = [
        ("constraint", EmissionConstraintWrap),
        ("eps_1", ctypes.c_float),
        ("eps_2", ctypes.c_float),
        ("tau", ctypes.c_float),
        ("k_max", ctypes.c_uint32),
        ("initial", ctypes.POINTER(ctypes.c_float)),
        ("initial_len", ctypes.c_uint32),
    ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, LMOption) and self._fields_ == other._fields_  # pragma: no cover


class GSOption(ctypes.Structure):
    _fields_ = [("constraint", EmissionConstraintWrap), ("repeat", ctypes.c_uint32)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GSOption) and self._fields_ == other._fields_  # pragma: no cover


class NaiveOption(ctypes.Structure):
    _fields_ = [("constraint", EmissionConstraintWrap)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NaiveOption) and self._fields_ == other._fields_  # pragma: no cover


class BackendPtr(ctypes.Structure):
    _fields_ = [("value", ctypes.c_void_p)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, BackendPtr) and self._fields_ == other._fields_  # pragma: no cover


class GreedyOption(ctypes.Structure):
    _fields_ = [("constraint", EmissionConstraintWrap), ("phase_div", ctypes.c_uint8)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GreedyOption) and self._fields_ == other._fields_  # pragma: no cover


class GSPATOption(ctypes.Structure):
    _fields_ = [("constraint", EmissionConstraintWrap), ("repeat", ctypes.c_uint32)]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, GSPATOption) and self._fields_ == other._fields_  # pragma: no cover


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi_gain_holo{bin_ext}"))

        self.dll.AUTDGainHoloConstraintNormalize.argtypes = []
        self.dll.AUTDGainHoloConstraintNormalize.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintUniform.argtypes = [EmitIntensity]
        self.dll.AUTDGainHoloConstraintUniform.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintMultiply.argtypes = [ctypes.c_float]
        self.dll.AUTDGainHoloConstraintMultiply.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloConstraintClamp.argtypes = [EmitIntensity, EmitIntensity]
        self.dll.AUTDGainHoloConstraintClamp.restype = EmissionConstraintWrap

        self.dll.AUTDGainHoloGreedySphere.argtypes = [ctypes.POINTER(Point3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, GreedyOption]
        self.dll.AUTDGainHoloGreedySphere.restype = GainPtr

        self.dll.AUTDGainGreedyIsDefault.argtypes = [GreedyOption]
        self.dll.AUTDGainGreedyIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloGSSphere.argtypes = [BackendPtr, ctypes.POINTER(Point3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, GSOption]
        self.dll.AUTDGainHoloGSSphere.restype = GainPtr

        self.dll.AUTDGainGSIsDefault.argtypes = [GSOption]
        self.dll.AUTDGainGSIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloGSPATSphere.argtypes = [BackendPtr, ctypes.POINTER(Point3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, GSPATOption]
        self.dll.AUTDGainHoloGSPATSphere.restype = GainPtr

        self.dll.AUTDGainGSPATIsDefault.argtypes = [GSPATOption]
        self.dll.AUTDGainGSPATIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloSPLToPascal.argtypes = [ctypes.c_float]
        self.dll.AUTDGainHoloSPLToPascal.restype = ctypes.c_float

        self.dll.AUTDGainHoloPascalToSPL.argtypes = [ctypes.c_float]
        self.dll.AUTDGainHoloPascalToSPL.restype = ctypes.c_float

        self.dll.AUTDGainHoloLMSphere.argtypes = [BackendPtr, ctypes.POINTER(Point3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, LMOption]
        self.dll.AUTDGainHoloLMSphere.restype = GainPtr

        self.dll.AUTDGainLMIsDefault.argtypes = [LMOption]
        self.dll.AUTDGainLMIsDefault.restype = ctypes.c_bool

        self.dll.AUTDGainHoloNaiveSphere.argtypes = [BackendPtr, ctypes.POINTER(Point3), ctypes.POINTER(ctypes.c_float), ctypes.c_uint32, NaiveOption]
        self.dll.AUTDGainHoloNaiveSphere.restype = GainPtr

        self.dll.AUTDGainNaiveIsDefault.argtypes = [NaiveOption]
        self.dll.AUTDGainNaiveIsDefault.restype = ctypes.c_bool

        self.dll.AUTDNalgebraBackendSphere.argtypes = []
        self.dll.AUTDNalgebraBackendSphere.restype = BackendPtr

        self.dll.AUTDDeleteNalgebraBackendSphere.argtypes = [BackendPtr]
        self.dll.AUTDDeleteNalgebraBackendSphere.restype = None

    def gain_holo_constraint_normalize(self) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintNormalize()

    def gain_holo_constraint_uniform(self, intensity: EmitIntensity) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintUniform(intensity)

    def gain_holo_constraint_multiply(self, v: float) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintMultiply(v)

    def gain_holo_constraint_clamp(self, min_v: EmitIntensity, max_v: EmitIntensity) -> EmissionConstraintWrap:
        return self.dll.AUTDGainHoloConstraintClamp(min_v, max_v)

    def gain_holo_greedy_sphere(self, points: ctypes.Array[Point3], amps: ctypes.Array[ctypes.c_float], size: int, option: GreedyOption) -> GainPtr:
        return self.dll.AUTDGainHoloGreedySphere(points, amps, size, option)

    def gain_greedy_is_default(self, option: GreedyOption) -> ctypes.c_bool:
        return self.dll.AUTDGainGreedyIsDefault(option)

    def gain_holo_gs_sphere(
        self,
        backend: BackendPtr,
        points: ctypes.Array[Point3],
        amps: ctypes.Array[ctypes.c_float],
        size: int,
        option: GSOption,
    ) -> GainPtr:
        return self.dll.AUTDGainHoloGSSphere(backend, points, amps, size, option)

    def gain_gs_is_default(self, option: GSOption) -> ctypes.c_bool:
        return self.dll.AUTDGainGSIsDefault(option)

    def gain_holo_gspat_sphere(
        self,
        backend: BackendPtr,
        points: ctypes.Array[Point3],
        amps: ctypes.Array[ctypes.c_float],
        size: int,
        option: GSPATOption,
    ) -> GainPtr:
        return self.dll.AUTDGainHoloGSPATSphere(backend, points, amps, size, option)

    def gain_gspat_is_default(self, option: GSPATOption) -> ctypes.c_bool:
        return self.dll.AUTDGainGSPATIsDefault(option)

    def gain_holo_spl_to_pascal(self, value: float) -> ctypes.c_float:
        return self.dll.AUTDGainHoloSPLToPascal(value)

    def gain_holo_pascal_to_spl(self, value: float) -> ctypes.c_float:
        return self.dll.AUTDGainHoloPascalToSPL(value)

    def gain_holo_lm_sphere(
        self,
        backend: BackendPtr,
        points: ctypes.Array[Point3],
        amps: ctypes.Array[ctypes.c_float],
        size: int,
        option: LMOption,
    ) -> GainPtr:
        return self.dll.AUTDGainHoloLMSphere(backend, points, amps, size, option)

    def gain_lm_is_default(self, option: LMOption) -> ctypes.c_bool:
        return self.dll.AUTDGainLMIsDefault(option)

    def gain_holo_naive_sphere(
        self,
        backend: BackendPtr,
        points: ctypes.Array[Point3],
        amps: ctypes.Array[ctypes.c_float],
        size: int,
        option: NaiveOption,
    ) -> GainPtr:
        return self.dll.AUTDGainHoloNaiveSphere(backend, points, amps, size, option)

    def gain_naive_is_default(self, option: NaiveOption) -> ctypes.c_bool:
        return self.dll.AUTDGainNaiveIsDefault(option)

    def nalgebra_backend_sphere(self) -> BackendPtr:
        return self.dll.AUTDNalgebraBackendSphere()

    def delete_nalgebra_backend_sphere(self, backend: BackendPtr) -> None:
        return self.dll.AUTDDeleteNalgebraBackendSphere(backend)



from typing import Self

from pyautd3.native_methods.autd3 import DcSysTime as _DcSysTime
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.utils import Duration


class DcSysTime:
    _inner: _DcSysTime

    def __new__(cls: type["DcSysTime"]) -> "DcSysTime":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["DcSysTime"], inner: _DcSysTime) -> "DcSysTime":
        ins = super().__new__(cls)
        ins._inner = inner
        return ins

    def sys_time(self: Self) -> int:
        return int(self._inner.dc_sys_time)

    @staticmethod
    def now() -> "DcSysTime":
        return DcSysTime.__private_new__(Base().dc_sys_time_now())

    def __add__(self: Self, other: Duration) -> "DcSysTime":
        sys_time = self.sys_time() + other.as_nanos()
        inner = _DcSysTime()
        inner.dc_sys_time = sys_time
        return DcSysTime.__private_new__(inner)

    def __sub__(self: Self, other: Duration) -> "DcSysTime":
        sys_time = self.sys_time() - other.as_nanos()
        inner = _DcSysTime()
        inner.dc_sys_time = sys_time
        return DcSysTime.__private_new__(inner)

    ZERO: "DcSysTime" = None  # type: ignore[assignment]


DcSysTime.ZERO = DcSysTime.__private_new__(_DcSysTime(0x00))



import contextlib
import os
import os.path
import platform
import sys
from pathlib import Path

from .autd3capi import NativeMethods as Base
from .autd3capi_gain_holo import NativeMethods as GainHolo
from .autd3capi_link_simulator import NativeMethods as LinkSimulator
from .autd3capi_link_twincat import NativeMethods as LinkTwincAT
from .autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile

_PLATFORM = platform.system()
_PREFIX = ""
_BIN_EXT = ""
if _PLATFORM == "Windows":
    _BIN_EXT = ".dll"
elif _PLATFORM == "Darwin":
    _PREFIX = "lib"
    _BIN_EXT = ".dylib"
elif _PLATFORM == "Linux":
    _PREFIX = "lib"
    _BIN_EXT = ".so"
else:
    err = "Not supported OS"
    raise ImportError(err)

_LIB_PATH = Path(__file__).parent.parent / "bin"

Base().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
GainHolo().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
ModulationAudioFile().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
LinkSimulator().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)
if sys.platform == "win32":
    with contextlib.suppress(FileNotFoundError):
        os.add_dll_directory("C:\\TwinCAT\\Common64")
with contextlib.suppress(Exception):
    LinkTwincAT().init_dll(_LIB_PATH, _PREFIX, _BIN_EXT)



from typing import Self

from pyautd3.driver.defined import Freq
from pyautd3.driver.defined.freq import Hz
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SamplingConfigWrap
from pyautd3.native_methods.utils import _validate_duration, _validate_f32, _validate_sampling_config, _validate_u16
from pyautd3.utils import Duration


class SamplingConfig:
    _inner: SamplingConfigWrap

    def __init__(self: Self, value: "SamplingConfigWrap | SamplingConfig | int | Freq[int] | Freq[float] | Duration") -> None:
        match value:
            case int():
                if value > 0xFFFF:  # noqa: PLR2004
                    raise ValueError
                self._inner = _validate_sampling_config(Base().sampling_config_from_division(value))
            case SamplingConfigWrap():
                self._inner = value
            case SamplingConfig():
                self._inner = value._inner
            case Freq():
                match value.hz():
                    case float() as v:
                        self._inner = Base().sampling_config_from_freq(v)
                    case _:
                        raise TypeError
            case Duration():
                self._inner = Base().sampling_config_from_period(value._inner)
            case _:
                raise TypeError

    def into_nearest(self: Self) -> "SamplingConfig":
        return SamplingConfig(Base().sampling_config_into_nearest(self._inner))

    @property
    def division(self: Self) -> int:
        return _validate_u16(Base().sampling_config_division(self._inner))

    def freq(self: Self) -> Freq[float]:
        return _validate_f32(Base().sampling_config_freq(self._inner)) * Hz

    def period(self: Self) -> Duration:
        return Duration.__private_new__(_validate_duration(Base().sampling_config_period(self._inner)))

    def __eq__(self: Self, value: object) -> bool:
        return isinstance(value, SamplingConfig) and bool(Base().sampling_config_eq(self._inner, value._inner))

    FREQ_40K: "SamplingConfig" = None  # type: ignore[assignment]
    FREQ_4K: "SamplingConfig" = None  # type: ignore[assignment]


SamplingConfig.FREQ_40K = SamplingConfig(40000.0 * Hz)
SamplingConfig.FREQ_4K = SamplingConfig(4000.0 * Hz)



from typing import Self

from pyautd3.native_methods.autd3capi_gain_holo import NativeMethods as GainHolo


class Amplitude:
    _value: float

    def __new__(cls: type["Amplitude"]) -> "Amplitude":
        raise NotImplementedError

    @classmethod
    def __private_new__(cls: type["Amplitude"], value: float) -> "Amplitude":
        ins = super().__new__(cls)
        ins._value = value
        return ins

    @staticmethod
    def new_pascal(value: float) -> "Amplitude":
        return Amplitude.__private_new__(value)

    @staticmethod
    def new_spl(value: float) -> "Amplitude":
        return Amplitude.__private_new__(float(GainHolo().gain_holo_spl_to_pascal(value)))

    def pascal(self: Self) -> float:
        return self._value

    def spl(self: Self) -> float:
        return float(GainHolo().gain_holo_pascal_to_spl(self._value))

    class _UnitPascal:
        def __new__(cls: type["Amplitude._UnitPascal"]) -> "Amplitude._UnitPascal":
            raise NotImplementedError

        @classmethod
        def __private_new__(cls: type["Amplitude._UnitPascal"]) -> "Amplitude._UnitPascal":
            return super().__new__(cls)

        def __rmul__(self: Self, other: float) -> "Amplitude":
            return Amplitude.new_pascal(other)

    class _UnitSPL:
        def __new__(cls: type["Amplitude._UnitSPL"]) -> "Amplitude._UnitSPL":
            raise NotImplementedError

        @classmethod
        def __private_new__(cls: type["Amplitude._UnitSPL"]) -> "Amplitude._UnitSPL":
            return super().__new__(cls)

        def __rmul__(self: Self, other: float) -> "Amplitude":
            return Amplitude.new_spl(other)


Pa: Amplitude._UnitPascal = Amplitude._UnitPascal.__private_new__()
dB: Amplitude._UnitSPL = Amplitude._UnitSPL.__private_new__()  # noqa: N816



import ctypes
import threading
from pathlib import Path

from pyautd3.native_methods.autd3capi_driver import ResultLink, ResultStatus


class Singleton(type):
    _instances = {}  # type: ignore[var-annotated]
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with cls._lock:
                if cls not in cls._instances:  # pragma: no cover
                    cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class NativeMethods(metaclass=Singleton):
    def init_dll(self, bin_location: Path, bin_prefix: str, bin_ext: str) -> None:
        self.dll = ctypes.CDLL(str(bin_location / f"{bin_prefix}autd3capi_link_simulator{bin_ext}"))

        self.dll.AUTDLinkSimulatorTracingInit.argtypes = []
        self.dll.AUTDLinkSimulatorTracingInit.restype = None

        self.dll.AUTDLinkSimulatorTracingInitWithFile.argtypes = [ctypes.c_char_p]
        self.dll.AUTDLinkSimulatorTracingInitWithFile.restype = ResultStatus

        self.dll.AUTDLinkSimulator.argtypes = [ctypes.c_char_p]
        self.dll.AUTDLinkSimulator.restype = ResultLink

    def link_simulator_tracing_init(self) -> None:
        return self.dll.AUTDLinkSimulatorTracingInit()

    def link_simulator_tracing_init_with_file(self, path: bytes) -> ResultStatus:
        return self.dll.AUTDLinkSimulatorTracingInitWithFile(path)

    def link_simulator(self, addr: bytes) -> ResultLink:
        return self.dll.AUTDLinkSimulator(addr)



import ctypes
from collections.abc import Callable, Iterable
from types import TracebackType
from typing import Generic, Self, TypeVar

import numpy as np

from pyautd3.autd_error import InvalidDatagramTypeError
from pyautd3.controller.sleeper import SpinSleeper, StdSleeper, WaitableSleeper
from pyautd3.driver.autd3_device import AUTD3
from pyautd3.driver.datagram import Datagram
from pyautd3.driver.firmware.fpga import FPGAState
from pyautd3.driver.firmware_version import FirmwareInfo
from pyautd3.driver.geometry import Device, Geometry
from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3 import ParallelMode
from pyautd3.native_methods.autd3capi import ControllerPtr, SenderPtr
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SenderOption as SenderOption_
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GeometryPtr
from pyautd3.native_methods.structs import Point3, Quaternion
from pyautd3.native_methods.utils import _validate_ptr, _validate_status
from pyautd3.utils import Duration
from pyautd3.utils.duration import into_option_duration

K = TypeVar("K")
L = TypeVar("L", bound=Link)


class SenderOption:
    send_interval: Duration
    receive_interval: Duration
    timeout: Duration | None
    parallel: ParallelMode
    sleeper: StdSleeper | SpinSleeper | WaitableSleeper

    def __init__(
        self: Self,
        *,
        send_interval: Duration | None = None,
        receive_interval: Duration | None = None,
        timeout: Duration | None = None,
        parallel: ParallelMode = ParallelMode.Auto,
        sleeper: StdSleeper | SpinSleeper | WaitableSleeper | None = None,
    ) -> None:
        self.send_interval = send_interval or Duration.from_millis(1)
        self.receive_interval = receive_interval or Duration.from_millis(1)
        self.timeout = timeout
        self.parallel = parallel
        self.sleeper = sleeper or SpinSleeper()

    def _inner(self: Self) -> SenderOption_:
        return SenderOption_(
            self.send_interval._inner,
            self.receive_interval._inner,
            into_option_duration(self.timeout),
            self.parallel,
            self.sleeper._inner(),
        )


class Sender:
    _ptr: SenderPtr
    _geometry: Geometry

    def __init__(self: Self, ptr: SenderPtr, geometry: Geometry) -> None:
        self._ptr = ptr
        self._geometry = geometry

    def send(
        self: Self,
        d: Datagram | tuple[Datagram, Datagram],
    ) -> None:
        match d:
            case Datagram():
                result = Base().sender_send(self._ptr, d._datagram_ptr(self._geometry))
            case (Datagram(), Datagram()):
                (d1, d2) = d
                d_tuple = Base().datagram_tuple(d1._datagram_ptr(self._geometry), d2._datagram_ptr(self._geometry))
                result = Base().sender_send(self._ptr, d_tuple)
            case _:
                raise InvalidDatagramTypeError
        _validate_status(result)

    def group_send(
        self: Self,
        key_map: Callable[[Device], K | None],
        data_map: dict[K, Datagram | tuple[Datagram, Datagram]],
    ) -> None:
        keymap: dict[K, int] = {}
        datagrams: np.ndarray = np.ndarray(len(data_map), dtype=DatagramPtr)
        keys = np.arange(len(data_map), dtype=np.int32)

        def f_native(_context: ctypes.c_void_p, geometry_ptr: GeometryPtr, dev_idx: int) -> int:
            dev = Device(dev_idx, geometry_ptr)
            key = key_map(dev)
            return keymap[key] if key is not None else -1

        f_native_ = ctypes.CFUNCTYPE(ctypes.c_int32, ctypes.c_void_p, GeometryPtr, ctypes.c_uint16)(f_native)

        for k, (key, d) in enumerate(data_map.items()):
            match d:
                case Datagram():
                    ptr = d._datagram_ptr(self._geometry)
                    datagrams[k]["value"] = ptr.value
                case (Datagram(), Datagram()):
                    (d1, d2) = d
                    ptr = Base().datagram_tuple(
                        d1._datagram_ptr(self._geometry),
                        d2._datagram_ptr(self._geometry),
                    )
                    datagrams[k]["value"] = ptr.value
                case _:
                    raise InvalidDatagramTypeError
            keymap[key] = k

        _validate_status(
            Base().controller_group(
                self._ptr,
                f_native_,  # type: ignore[arg-type]
                ctypes.c_void_p(0),
                self._geometry._geometry_ptr,
                keys.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),  # type: ignore[arg-type]
                datagrams.ctypes.data_as(ctypes.POINTER(DatagramPtr)),  # type: ignore[arg-type]
                len(keys),
            ),
        )


class Controller(Geometry, Generic[L]):
    _ptr: ControllerPtr
    _link: L
    _disposed: bool

    def __init__(self: Self, geometry: GeometryPtr, ptr: ControllerPtr, link: L) -> None:
        super().__init__(geometry)
        self._ptr = ptr
        self._link = link
        self._link._ptr = Base().link_get(self._ptr)
        self._disposed = False

    def link(self: Self) -> L:
        return self._link

    def __del__(self: Self) -> None:
        self._dispose()

    def _dispose(self: Self) -> None:
        self.close()

    def __enter__(self: Self) -> "Controller[L]":
        return self

    def __exit__(
        self: Self,
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: TracebackType | None,
    ) -> None:
        self._dispose()

    def geometry(self: Self) -> Geometry:
        return self  # type: ignore[return-value]

    @staticmethod
    def open(devices: Iterable[AUTD3], link: L) -> "Controller[L]":
        return Controller.open_with_option(devices, link, SenderOption())

    @staticmethod
    def open_with_option(devices: Iterable[AUTD3], link: L, option: SenderOption) -> "Controller[L]":
        devices = list(devices)
        pos = np.fromiter((np.void(Point3(d.pos)) for d in devices), dtype=Point3)  # type: ignore[type-var,call-overload]
        rot = np.fromiter((np.void(Quaternion(d.rot)) for d in devices), dtype=Quaternion)  # type: ignore[type-var,call-overload]

        ptr = _validate_ptr(
            Base().controller_open(
                pos.ctypes.data_as(ctypes.POINTER(Point3)),  # type: ignore[arg-type]
                rot.ctypes.data_as(ctypes.POINTER(Quaternion)),  # type: ignore[arg-type]
                len(devices),
                link._resolve(),
                option._inner(),
            ),
        )
        geometry = Base().geometry(ptr)
        return Controller(geometry, ptr, link)

    def firmware_version(self: Self) -> list[FirmwareInfo]:
        handle = _validate_ptr(
            Base().controller_firmware_version_list_pointer(self._ptr),
        )

        def get_firmware_info(i: int) -> FirmwareInfo:
            sb = bytes(bytearray(256))
            Base().controller_firmware_version_get(handle, i, sb)
            info = sb.decode("utf-8").rstrip(" \t\r\n\0")
            return FirmwareInfo(info)

        res = list(map(get_firmware_info, range(self.num_devices())))
        Base().controller_firmware_version_list_pointer_delete(handle)
        return res

    def close(self: Self) -> None:
        if self._disposed:
            return
        self._disposed = True
        r = Base().controller_close(self._ptr)
        self._ptr.value = None
        _validate_status(r)

    def fpga_state(self: Self) -> list[FPGAState | None]:
        handle = _validate_ptr(Base().controller_fpga_state(self._ptr))

        def get_fpga_state(i: int) -> FPGAState | None:
            state = int(Base().controller_fpga_state_get(handle, i))
            return None if state == -1 else FPGAState(state)

        res = list(map(get_fpga_state, range(self.num_devices())))
        Base().controller_fpga_state_delete(handle)
        return res

    def sender(self: Self, option: SenderOption) -> Sender:
        return Sender(Base().sender(self._ptr, option._inner()), self.geometry())

    def send(
        self: Self,
        d: Datagram | tuple[Datagram, Datagram],
    ) -> None:
        self.sender(SenderOption()).send(d)

    def group_send(
        self: Self,
        key_map: Callable[[Device], K | None],
        data_map: dict[K, Datagram | tuple[Datagram, Datagram]],
    ) -> None:
        self.sender(SenderOption()).group_send(key_map, data_map)



from typing import Self

from pyautd3.driver.defined import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.utils import _validate_sampling_config
from pyautd3.utils import Duration


class FreqNearest:
    freq: Freq[float]

    def __init__(self: Self, freq: Freq[float]) -> None:
        self.freq = freq


class PeriodNearest:
    period: Duration

    def __init__(self: Self, period: Duration) -> None:
        self.period = period


def _sampling_config(config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest, n: int) -> SamplingConfig:
    match config:
        case SamplingConfig() as value:
            return value
        case Freq() as value:
            return SamplingConfig(_validate_sampling_config(Base().stm_config_from_freq(value.hz(), n)))
        case Duration() as value:
            return SamplingConfig(_validate_sampling_config(Base().stm_config_from_period(value._inner, n)))
        case FreqNearest() as value:
            return SamplingConfig(Base().stm_config_from_freq_nearest(value.freq.hz(), n))
        case PeriodNearest() as value:  # pragma: no cover
            return SamplingConfig(Base().stm_config_from_period_nearest(value.period._inner, n))



import ctypes
from abc import abstractmethod
from typing import Self

import numpy as np
from numpy.typing import ArrayLike

from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.native_methods.structs import Point3

# _pad is not required for python 3.12+?


class ControlPoint(ctypes.Structure):
    _fields_ = [  # noqa: RUF012
        ("_point", Point3),
        ("_offset", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, point: ArrayLike, phase_offset: Phase | None = None) -> None:
        super().__init__()
        self.point = np.array(point)
        self.phase_offset = phase_offset or Phase(0)

    @property
    def point(self: Self) -> np.ndarray:
        return self._point.ndarray()

    @point.setter
    def point(self: Self, value: ArrayLike) -> None:
        self._point = Point3(np.array(value))

    @property
    def phase_offset(self: Self) -> Phase:
        return Phase(self._offset)

    @phase_offset.setter
    def phase_offset(self: Self, value: Phase) -> None:
        self._offset = value.value


class IControlPoints:
    @staticmethod
    @abstractmethod
    def _value() -> int:
        pass


class ControlPoints1(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(self: Self, points: ArrayLike | ControlPoint, intensity: EmitIntensity | None = None) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array([self._point.point])

    @points.setter
    def points(self: Self, value: ArrayLike | ControlPoint) -> None:
        match value:
            case ControlPoint():
                self._point = value
            case _:
                self._point = ControlPoint(point=value)

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 1


class ControlPoints2(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array([self._point1.point, self._point2.point])  # type: ignore[union-attr]

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint()):
                self._point1, self._point2 = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2 = ControlPoint(point=value[0]), ControlPoint(point=value[1])

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 2


class ControlPoints3(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array([self._point1.point, self._point2.point, self._point3.point])  # type: ignore[union-attr]

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3 = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3 = ControlPoint(point=value[0]), ControlPoint(point=value[1]), ControlPoint(point=value[2])

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 3


class ControlPoints4(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array([self._point1.point, self._point2.point, self._point3.point, self._point4.point])  # type: ignore[union-attr]

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike] | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4 = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4 = (
                    ControlPoint(point=value[0]),
                    ControlPoint(point=value[1]),
                    ControlPoint(point=value[2]),
                    ControlPoint(point=value[3]),
                )

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 4


class ControlPoints5(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array([self._point1.point, self._point2.point, self._point3.point, self._point4.point, self._point5.point])  # type: ignore[union-attr]

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4, self._point5 = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5 = (
                    ControlPoint(point=value[0]),
                    ControlPoint(point=value[1]),
                    ControlPoint(point=value[2]),
                    ControlPoint(point=value[3]),
                    ControlPoint(point=value[4]),
                )

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 5


class ControlPoints6(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array(
            [
                self._point1.point,  # type: ignore[union-attr]
                self._point2.point,  # type: ignore[union-attr]
                self._point3.point,  # type: ignore[union-attr]
                self._point4.point,  # type: ignore[union-attr]
                self._point5.point,  # type: ignore[union-attr]
                self._point6.point,  # type: ignore[union-attr]
            ],
        )

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6 = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6 = (
                    ControlPoint(point=value[0]),
                    ControlPoint(point=value[1]),
                    ControlPoint(point=value[2]),
                    ControlPoint(point=value[3]),
                    ControlPoint(point=value[4]),
                    ControlPoint(point=value[5]),
                )

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 6


class ControlPoints7(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array(
            [
                self._point1.point,  # type: ignore[union-attr]
                self._point2.point,  # type: ignore[union-attr]
                self._point3.point,  # type: ignore[union-attr]
                self._point4.point,  # type: ignore[union-attr]
                self._point5.point,  # type: ignore[union-attr]
                self._point6.point,  # type: ignore[union-attr]
                self._point7.point,  # type: ignore[union-attr]
            ],
        )

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                (
                    self._point1,
                    self._point2,
                    self._point3,
                    self._point4,
                    self._point5,
                    self._point6,
                    self._point7,
                ) = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6, self._point7 = (
                    ControlPoint(point=value[0]),
                    ControlPoint(point=value[1]),
                    ControlPoint(point=value[2]),
                    ControlPoint(point=value[3]),
                    ControlPoint(point=value[4]),
                    ControlPoint(point=value[5]),
                    ControlPoint(point=value[6]),
                )

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 7


class ControlPoints8(ctypes.Structure, IControlPoints):
    _fields_ = [  # noqa: RUF012
        ("_point1", ControlPoint),
        ("_point2", ControlPoint),
        ("_point3", ControlPoint),
        ("_point4", ControlPoint),
        ("_point5", ControlPoint),
        ("_point6", ControlPoint),
        ("_point7", ControlPoint),
        ("_point8", ControlPoint),
        ("_intensity", ctypes.c_uint8),
        ("_pad", ctypes.c_uint8 * 3),
    ]

    def __init__(
        self: Self,
        points: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
        intensity: EmitIntensity | None = None,
    ) -> None:
        super().__init__()
        self.points = points  # type: ignore[assignment]
        self.intensity = intensity or EmitIntensity.MAX

    @property
    def points(self: Self) -> np.ndarray:
        return np.array(
            [
                self._point1.point,  # type: ignore[union-attr]
                self._point2.point,  # type: ignore[union-attr]
                self._point3.point,  # type: ignore[union-attr]
                self._point4.point,  # type: ignore[union-attr]
                self._point5.point,  # type: ignore[union-attr]
                self._point6.point,  # type: ignore[union-attr]
                self._point7.point,  # type: ignore[union-attr]
                self._point8.point,  # type: ignore[union-attr]
            ],
        )

    @points.setter
    def points(
        self: Self,
        value: tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike, ArrayLike]
        | tuple[ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint, ControlPoint],
    ) -> None:
        match value:
            case (ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint(), ControlPoint()):
                (
                    self._point1,
                    self._point2,
                    self._point3,
                    self._point4,
                    self._point5,
                    self._point6,
                    self._point7,
                    self._point8,
                ) = value  # type: ignore[var-annotated]
            case _:
                self._point1, self._point2, self._point3, self._point4, self._point5, self._point6, self._point7, self._point8 = (
                    ControlPoint(point=value[0]),
                    ControlPoint(point=value[1]),
                    ControlPoint(point=value[2]),
                    ControlPoint(point=value[3]),
                    ControlPoint(point=value[4]),
                    ControlPoint(point=value[5]),
                    ControlPoint(point=value[6]),
                    ControlPoint(point=value[7]),
                )

    @property
    def intensity(self: Self) -> EmitIntensity:
        return EmitIntensity(self._intensity)

    @intensity.setter
    def intensity(self: Self, value: EmitIntensity) -> None:
        self._intensity = value.value

    @staticmethod
    def _value() -> int:
        return 8



from pyautd3.autd_error import AUTDError
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SamplingConfigWrap
from pyautd3.native_methods.autd3capi_driver import AUTDStatus, Duration, ResultDuration, ResultF32, ResultSamplingConfig, ResultStatus, ResultU16


def _to_null_terminated_utf8(s: str) -> bytes:
    return s.encode("utf-8") + b"\0"


def _validate_status(res: ResultStatus) -> int:
    if int(res.result) == AUTDStatus.AUTDErr:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return int(res.result)


def _validate_sampling_config(res: ResultSamplingConfig) -> SamplingConfigWrap:
    if int(res.err_len) != 0:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return res.result


def _validate_duration(res: ResultDuration) -> Duration:
    if int(res.err_len) != 0:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return res.result


def _validate_u16(res: ResultU16) -> int:
    if int(res.err_len) != 0:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return int(res.result)


def _validate_f32(res: ResultF32) -> float:
    if int(res.err_len) != 0:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return float(res.result)


def _validate_ptr(res):
    if res.result.value is None:
        err = bytes(bytearray(int(res.err_len)))
        Base().get_err(res.err, err)
        raise AUTDError(err)
    return res.result


class ConstantADT(type):
    _initialized = False

    def __setattr__(cls, name, value) -> None:
        if cls._initialized:  # pragma: no cover
            if name in cls.__dict__:  # pragma: no cover
                err = f"Do not assign value to {name}"
                raise ValueError(err)  # pragma: no cover
            err = f"Do not add new member to {cls}"
            raise AttributeError(err)  # pragma: no cover
        super().__setattr__(name, value)

    def __init__(cls, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        cls._initialized = True



from typing import Self

from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.firmware.fpga.emit_intensity import EmitIntensity
from pyautd3.driver.firmware.fpga.phase import Phase
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import GainPtr


class Uniform(Gain):
    intensity: EmitIntensity
    phase: Phase

    def __init__(self: Self, intensity: EmitIntensity, phase: Phase) -> None:
        super().__init__()
        self.intensity = intensity
        self.phase = phase

    def _gain_ptr(self: Self, _: Geometry) -> GainPtr:
        return Base().gain_uniform(self.intensity._inner(), self.phase._inner())



from typing import Self

from pyautd3.native_methods.autd3 import Segment


class FPGAState:
    _state: int

    def __init__(self: Self, info: int) -> None:
        self._state = info

    def is_thermal_assert(self: Self) -> bool:
        return (self._state & (1 << 0)) != 0

    def current_mod_segment(self: Self) -> Segment:
        match int(self._state) & (1 << 1):
            case 0:
                return Segment.S0
            case _:
                return Segment.S1

    def current_stm_segment(self: Self) -> Segment | None:
        if not self.is_stm_mode():
            return None

        match int(self._state) & (1 << 2):
            case 0:
                return Segment.S0
            case _:
                return Segment.S1

    def current_gain_segment(self: Self) -> Segment | None:
        if not self.is_gain_mode():
            return None

        match int(self._state) & (1 << 2):
            case 0:
                return Segment.S0
            case _:
                return Segment.S1

    def is_gain_mode(self: Self) -> bool:
        return (int(self._state) & (1 << 3)) != 0

    def is_stm_mode(self: Self) -> bool:
        return not self.is_gain_mode()



from typing import Self

from pyautd3.driver.link import Link
from pyautd3.native_methods.autd3capi_driver import LinkPtr
from pyautd3.native_methods.autd3capi_link_twincat import NativeMethods as LinkTwinCAT
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr


class TwinCAT(Link):
    def __init__(self: Self) -> None:
        super().__init__()

    def _resolve(self: Self) -> LinkPtr:
        return _validate_ptr(LinkTwinCAT().link_twin_cat())  # pragma: no cover


class RemoteTwinCATOption:
    server_ip: str
    client_ams_net_id: str

    def __init__(self: Self, *, server_ip: str = "", client_ams_net_id: str = "") -> None:
        self.server_ip = server_ip
        self.client_ams_net_id = client_ams_net_id


class RemoteTwinCAT(Link):
    server_ams_net_id: str
    option: RemoteTwinCATOption

    def __init__(self: Self, server_ams_net_id: str, option: RemoteTwinCATOption) -> None:
        super().__init__()
        self.server_ams_net_id = server_ams_net_id
        self.option = option

    def _resolve(self: Self) -> LinkPtr:
        return _validate_ptr(  # pragma: no cover
            LinkTwinCAT().link_remote_twin_cat(
                _to_null_terminated_utf8(self.server_ams_net_id),
                _to_null_terminated_utf8(self.option.server_ip),
                _to_null_terminated_utf8(self.option.client_ams_net_id),
            ),
        )



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.utils import Duration


class Custom(Modulation):
    buffer: np.ndarray
    config: SamplingConfig | Freq[int] | Freq[float] | Duration

    def __init__(self: Self, buffer: Iterable[int], sampling_config: SamplingConfig | Freq[int] | Freq[float] | Duration) -> None:
        super().__init__()
        self.buffer = np.fromiter(buffer, dtype=np.uint8)
        self.config = sampling_config

    def _modulation_ptr(self: Self) -> ModulationPtr:
        return Base().modulation_custom(
            self.buffer.ctypes.data_as(ctypes.POINTER(ctypes.c_uint8)),  # type: ignore[arg-type]
            len(self.buffer),
            SamplingConfig(self.config)._inner,
        )



from typing import Self

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import Segment
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, TransitionModeWrap


class SwapSegment:
    def __new__(cls: type["SwapSegment"]) -> "SwapSegment":
        raise NotImplementedError

    class Gain(Datagram):
        _segment: Segment
        _transition_mode: TransitionModeWrap

        def __init__(self: Self, segment: Segment, transition_mode: TransitionModeWrap) -> None:
            super().__init__()
            self._segment = segment
            self._transition_mode = transition_mode

        def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
            return Base().datagram_swap_segment_gain(self._segment, self._transition_mode)

    class Modulation(Datagram):
        _segment: Segment
        _transition_mode: TransitionModeWrap

        def __init__(self: Self, segment: Segment, transition_mode: TransitionModeWrap) -> None:
            super().__init__()
            self._segment = segment
            self._transition_mode = transition_mode

        def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
            return Base().datagram_swap_segment_modulation(self._segment, self._transition_mode)

    class FociSTM(Datagram):
        _segment: Segment
        _transition_mode: TransitionModeWrap

        def __init__(self: Self, segment: Segment, transition_mode: TransitionModeWrap) -> None:
            super().__init__()
            self._segment = segment
            self._transition_mode = transition_mode

        def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
            return Base().datagram_swap_segment_foci_stm(self._segment, self._transition_mode)

    class GainSTM(Datagram):
        _segment: Segment
        _transition_mode: TransitionModeWrap

        def __init__(self: Self, segment: Segment, transition_mode: TransitionModeWrap) -> None:
            super().__init__()
            self._segment = segment
            self._transition_mode = transition_mode

        def _datagram_ptr(self: Self, _: Geometry) -> DatagramPtr:
            return Base().datagram_swap_segment_gain_stm(self._segment, self._transition_mode)



from pathlib import Path
from typing import Self

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.native_methods.autd3capi_driver import ModulationPtr
from pyautd3.native_methods.autd3capi_modulation_audio_file import NativeMethods as ModulationAudioFile
from pyautd3.native_methods.utils import _to_null_terminated_utf8, _validate_ptr
from pyautd3.utils import Duration


class CsvOption:
    delimiter: str

    def __init__(self: Self, *, delimiter: str = ",") -> None:
        self.delimiter = delimiter


class Csv(Modulation):
    path: Path
    config: SamplingConfig | Freq[int] | Freq[float] | Duration
    option: CsvOption

    def __init__(self: Self, path: Path, sampling_config: SamplingConfig | Freq[int] | Freq[float] | Duration, option: CsvOption) -> None:
        super().__init__()
        self.path = path
        self.config = sampling_config
        self.option = option

    def _modulation_ptr(self: Self) -> ModulationPtr:
        delim = self.option.delimiter.encode("utf-8")
        path = _to_null_terminated_utf8(str(self.path))
        return _validate_ptr(ModulationAudioFile().modulation_audio_file_csv(path, SamplingConfig(self.config)._inner, delim[0]))



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np

from pyautd3.driver.datagram.datagram import Datagram
from pyautd3.driver.datagram.gain import Gain
from pyautd3.driver.datagram.stm.stm_sampling_config import FreqNearest, PeriodNearest, _sampling_config
from pyautd3.driver.datagram.with_loop_behavior import DatagramL
from pyautd3.driver.datagram.with_segment import DatagramS
from pyautd3.driver.defined.freq import Freq
from pyautd3.driver.firmware.fpga.sampling_config import SamplingConfig
from pyautd3.driver.firmware.fpga.transition_mode import TransitionMode
from pyautd3.driver.geometry import Geometry
from pyautd3.native_methods.autd3 import GainSTMMode, Segment
from pyautd3.native_methods.autd3 import GainSTMOption as GainSTMOption_
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi_driver import DatagramPtr, GainPtr, GainSTMPtr, LoopBehavior, TransitionModeWrap
from pyautd3.utils import Duration

__all__ = []  # type: ignore[var-annotated]


class GainSTMOption:
    mode: GainSTMMode

    def __init__(self: Self, *, mode: GainSTMMode = GainSTMMode.PhaseIntensityFull) -> None:
        self.mode = mode

    def _inner(self: Self) -> GainSTMOption_:
        return GainSTMOption_(self.mode)


class GainSTM(DatagramS[GainSTMPtr], DatagramL[GainSTMPtr], Datagram):
    gains: list[Gain]
    config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest
    option: GainSTMOption

    @classmethod
    def __private_new__(
        cls: type["GainSTM"],
        gains: Iterable[Gain],
        config: SamplingConfig | Freq[float] | Duration | FreqNearest | PeriodNearest,
        option: GainSTMOption,
    ) -> "GainSTM":
        ins = super().__new__(cls)
        ins.gains = list(gains)
        ins.config = config
        ins.option = option
        return ins

    def __init__(
        self: Self,
        gains: Iterable[Gain],
        config: SamplingConfig | Freq[float] | Duration,
        option: GainSTMOption,
    ) -> None:
        self.gains = list(gains)
        self.config = config
        self.option = option

    def into_nearest(self: Self) -> "GainSTM":
        match self.config:
            case Freq() as freq:
                return GainSTM.__private_new__(self.gains, FreqNearest(freq), self.option)
            case Duration() as period:
                return GainSTM.__private_new__(self.gains, PeriodNearest(period), self.option)
            case _:
                raise TypeError

    def _raw_ptr(self: Self, geometry: Geometry) -> GainSTMPtr:
        gains: np.ndarray = np.ndarray(len(self.gains), dtype=GainPtr)
        for i, g in enumerate(self.gains):
            gains[i]["value"] = g._gain_ptr(geometry).value
        return Base().stm_gain(
            self.sampling_config()._inner,
            gains.ctypes.data_as(ctypes.POINTER(GainPtr)),  # type: ignore[arg-type]
            len(gains),
            self.option._inner(),
        )

    def _into_segment(
        self: Self,
        ptr: GainSTMPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
    ) -> DatagramPtr:
        return Base().stm_gain_into_datagram_with_segment(ptr, segment, transition_mode or TransitionMode.NONE)

    def _into_loop_behavior(
        self: Self,
        ptr: GainSTMPtr,
        segment: Segment,
        transition_mode: TransitionModeWrap | None,
        loop_behavior: LoopBehavior,
    ) -> DatagramPtr:
        return Base().stm_gain_into_datagram_with_loop_behavior(ptr, segment, transition_mode or TransitionMode.NONE, loop_behavior)

    def _datagram_ptr(self: Self, geometry: Geometry) -> DatagramPtr:
        return Base().stm_gain_into_datagram(self._raw_ptr(geometry))

    def sampling_config(self: Self) -> SamplingConfig:
        return _sampling_config(self.config, len(self.gains))



from abc import ABCMeta, abstractmethod
from typing import Self, TypeVar

from pyautd3.native_methods.autd3capi_driver import LinkPtr

__all__ = []  # type: ignore[var-annotated]

L = TypeVar("L", bound="Link")


class Link(metaclass=ABCMeta):
    _ptr: LinkPtr

    def __init__(self: Self) -> None:
        self._ptr = LinkPtr()

    @abstractmethod
    def _resolve(self: Self) -> LinkPtr:
        pass



import ctypes
from collections.abc import Iterable
from typing import Self

import numpy as np

from pyautd3.driver.datagram.modulation import Modulation
from pyautd3.modulation.sine import Sine, SineMode
from pyautd3.native_methods.autd3capi import FourierOption as FourierOption_
from pyautd3.native_methods.autd3capi import NativeMethods as Base
from pyautd3.native_methods.autd3capi import SineOption as SineOption_
from pyautd3.native_methods.autd3capi_driver import ModulationPtr


class FourierOption:
    scale_factor: float | None
    clamp: bool
    offset: int

    def __init__(
        self: Self,
        *,
        scale_factor: float | None = None,
        clamp: bool = False,
        offset: int = 0,
    ) -> None:
        self.scale_factor = scale_factor
        self.clamp = clamp or False
        self.offset = offset or 0

    def _inner(self: Self) -> FourierOption_:
        return FourierOption_(
            self.scale_factor is not None,
            self.scale_factor or float("nan"),
            self.clamp,
            self.offset,
        )


class Fourier(Modulation):
    components: list[Sine]
    option: FourierOption

    def __init__(self: Self, components: Iterable[Sine], option: FourierOption) -> None:
        super().__init__()
        self.components = list(components)
        self.option = option

    def _modulation_ptr(self: Self) -> ModulationPtr:
        size = len(self.components)
        option = self.option._inner()

        sine_option = np.fromiter((np.void(m.option._inner()) for m in self.components), dtype=SineOption_)  # type: ignore[type-var,call-overload]
        match self.components[0]._mode:
            case SineMode.Exact:
                sine_freq = np.fromiter((m.freq.hz() for m in self.components), dtype=np.uint32)
                return Base().modulation_fourier_exact(
                    sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),  # type: ignore[arg-type]
                    sine_option.ctypes.data_as(ctypes.POINTER(SineOption_)),  # type: ignore[arg-type]
                    size,
                    option,
                )
            case SineMode.ExactFloat:
                sine_freq = np.fromiter((m.freq.hz() for m in self.components), dtype=np.float32)
                return Base().modulation_fourier_exact_float(
                    sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                    sine_option.ctypes.data_as(ctypes.POINTER(SineOption_)),  # type: ignore[arg-type]
                    size,
                    option,
                )
            case SineMode.Nearest:  # pragma: no cover
                sine_freq = np.fromiter((m.freq.hz() for m in self.components), dtype=np.float32)
                return Base().modulation_fourier_nearest(
                    sine_freq.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),  # type: ignore[arg-type]
                    sine_option.ctypes.data_as(ctypes.POINTER(SineOption_)),  # type: ignore[arg-type]
                    size,
                    option,
                )
